@@ -23,11 +23,12 @@
 #include "box.h"
 #include "security_limits.h"
 #include "nclx.h"
-#include "jpeg.h"
-#include "jpeg2000.h"
-#include "hevc.h"
-#include "mask_image.h"
-#include "vvc.h"
+#include "codecs/jpeg.h"
+#include "codecs/jpeg2000.h"
+#include "codecs/hevc.h"
+#include "codecs/mask_image.h"
+#include "codecs/vvc.h"
+#include "codecs/avc.h"
 
 #include <iomanip>
 #include <utility>
@@ -36,9 +37,14 @@
 #include <cstring>
 #include <set>
 #include <cassert>
+#include <array>
 
 #if WITH_UNCOMPRESSED_CODEC
-#include "uncompressed_image.h"
+#include "codecs/uncompressed_box.h"
+#endif
+
+#ifndef M_PI
+#define M_PI 3.14159265358979323846
 #endif
 
 
@@ -192,13 +198,13 @@ std::string BoxHeader::get_type_string() const
     std::ostringstream sstr;
     sstr << std::hex;
     sstr << std::setfill('0');
-    sstr << std::setw(2);
 
     for (int i = 0; i < 16; i++) {
       if (i == 4 || i == 6 || i == 8 || i == 10) {
         sstr << '-';
       }
 
+      sstr << std::setw(2);
       sstr << ((int) m_uuid_type[i]);
     }
 
@@ -210,6 +216,23 @@ std::string BoxHeader::get_type_string() const
 }
 
 
+std::vector<uint8_t> BoxHeader::get_uuid_type() const
+{
+  if (m_type != fourcc("uuid")) {
+    return {};
+  }
+
+  return m_uuid_type;
+}
+
+
+void BoxHeader::set_uuid_type(const std::vector<uint8_t>& type)
+{
+  m_type = fourcc("uuid");
+  m_uuid_type = type;
+}
+
+
 Error BoxHeader::parse_header(BitstreamRange& range)
 {
   StreamReader::grow_status status;
@@ -391,12 +414,15 @@ Error Box::parse(BitstreamRange& range)
   }
   else {
     uint64_t content_size = get_box_size() - get_header_size();
-    if (range.prepare_read(content_size)) {
-      if (content_size > MAX_BOX_SIZE) {
-        return Error(heif_error_Invalid_input,
-                     heif_suberror_Invalid_box_size);
-      }
 
+    assert(MAX_BOX_SIZE <= SIZE_MAX);
+
+    if (content_size > MAX_BOX_SIZE) {
+      return Error(heif_error_Invalid_input,
+                   heif_suberror_Invalid_box_size);
+    }
+
+    if (range.prepare_read(static_cast<size_t>(content_size))) {
       range.get_istream()->seek_cur(get_box_size() - get_header_size());
     }
   }
@@ -520,6 +546,18 @@ Error Box::read(BitstreamRange& range, std::shared_ptr<Box>* result)
       box = std::make_shared<Box_grpl>();
       break;
 
+    case fourcc("pymd"):
+      box = std::make_shared<Box_pymd>();
+      break;
+
+    case fourcc("altr"):
+      box = std::make_shared<Box_EntityToGroup>();
+      break;
+
+    case fourcc("ster"):
+      box = std::make_shared<Box_ster>();
+      break;
+
     case fourcc("dinf"):
       box = std::make_shared<Box_dinf>();
       break;
@@ -564,6 +602,14 @@ Error Box::read(BitstreamRange& range, std::shared_ptr<Box>* result)
       box = std::make_shared<Box_mdcv>();
       break;
 
+    case fourcc("cmin"):
+      box = std::make_shared<Box_cmin>();
+      break;
+
+    case fourcc("cmex"):
+      box = std::make_shared<Box_cmex>();
+      break;
+
     case fourcc("udes"):
       box = std::make_shared<Box_udes>();
       break;
@@ -580,6 +626,14 @@ Error Box::read(BitstreamRange& range, std::shared_ptr<Box>* result)
     case fourcc("uncC"):
       box = std::make_shared<Box_uncC>();
       break;
+
+    case fourcc("cmpC"):
+      box = std::make_shared<Box_cmpC>();
+      break;
+
+    case fourcc("icbr"):
+      box = std::make_shared<Box_icbr>();
+      break;
 #endif
 
     // --- JPEG 2000
@@ -610,56 +664,87 @@ Error Box::read(BitstreamRange& range, std::shared_ptr<Box>* result)
       box = std::make_shared<Box_mskC>();
       break;
 
-    default:
+    // --- AVC (H.264)
+
+    case fourcc("avcC"):
+      box = std::make_shared<Box_avcC>();
+      break;
+
+    case fourcc("mdat"):
+      // avoid generating a 'Box_other'
       box = std::make_shared<Box>();
       break;
+
+    case fourcc("uuid"):
+      if (hdr.get_uuid_type() == std::vector<uint8_t>{0x22, 0xcc, 0x04, 0xc7, 0xd6, 0xd9, 0x4e, 0x07, 0x9d, 0x90, 0x4e, 0xb6, 0xec, 0xba, 0xf3, 0xa3}) {
+        box = std::make_shared<Box_cmin>();
+      }
+      else if (hdr.get_uuid_type() == std::vector<uint8_t>{0x43, 0x63, 0xe9, 0x14, 0x5b, 0x7d, 0x4a, 0xab, 0x97, 0xae, 0xbe, 0xa6, 0x98, 0x03, 0xb4, 0x34}) {
+        box = std::make_shared<Box_cmex>();
+      }
+      else {
+        box = std::make_shared<Box_other>(hdr.get_short_type());
+      }
+      break;
+
+    default:
+      box = std::make_shared<Box_other>(hdr.get_short_type());
+      break;
   }
 
   box->set_short_header(hdr);
 
-  if (hdr.has_fixed_box_size() && hdr.get_box_size() < hdr.get_header_size()) {
-    std::stringstream sstr;
-    sstr << "Box size (" << hdr.get_box_size() << " bytes) smaller than header size ("
-         << hdr.get_header_size() << " bytes)";
-
-    // Sanity check.
-    return Error(heif_error_Invalid_input,
-                 heif_suberror_Invalid_box_size,
-                 sstr.str());
-  }
-
-
   if (range.get_nesting_level() > MAX_BOX_NESTING_LEVEL) {
     return Error(heif_error_Memory_allocation_error,
                  heif_suberror_Security_limit_exceeded,
                  "Security limit for maximum nesting of boxes has been exceeded");
   }
 
-
   if (hdr.has_fixed_box_size()) {
-    auto status = range.wait_for_available_bytes(hdr.get_box_size() - hdr.get_header_size());
+    // Sanity checks
+    if (hdr.get_box_size() < hdr.get_header_size()) {
+      std::stringstream sstr;
+      sstr << "Box size (" << hdr.get_box_size() << " bytes) smaller than header size ("
+           << hdr.get_header_size() << " bytes)";
+
+      return {heif_error_Invalid_input,
+              heif_suberror_Invalid_box_size,
+              sstr.str()};
+    }
+
+    // this is >= 0 because of above condition
+    auto nBytes = static_cast<uint64_t>(hdr.get_box_size() - hdr.get_header_size());
+    if (nBytes > SIZE_MAX) {
+      return {heif_error_Memory_allocation_error,
+              heif_suberror_Invalid_box_size,
+              "Box size too large"};
+    }
+
+    // Security check: make sure that box size does not exceed int64 size.
+
+    if (hdr.get_box_size() > (uint64_t) std::numeric_limits<int64_t>::max()) {
+      return {heif_error_Invalid_input,
+              heif_suberror_Invalid_box_size};
+    }
+
+    // --- wait for data to arrive
+
+    auto status = range.wait_for_available_bytes(static_cast<size_t>(nBytes));
     if (status != StreamReader::size_reached) {
       // TODO: return recoverable error at timeout
-      return Error(heif_error_Invalid_input,
-                   heif_suberror_End_of_data);
+      return {heif_error_Invalid_input,
+              heif_suberror_End_of_data};
     }
   }
 
-  // Security check: make sure that box size does not exceed int64 size.
-
-  if (hdr.get_box_size() > (uint64_t) std::numeric_limits<int64_t>::max()) {
-    return Error(heif_error_Invalid_input,
-                 heif_suberror_Invalid_box_size);
-  }
-
-  int64_t box_size = static_cast<int64_t>(hdr.get_box_size());
+  auto box_size = static_cast<int64_t>(hdr.get_box_size());
   int64_t box_size_without_header = hdr.has_fixed_box_size() ? (box_size - hdr.get_header_size()) : (int64_t)range.get_remaining_bytes();
 
   // Box size may not be larger than remaining bytes in parent box.
 
   if ((int64_t)range.get_remaining_bytes() < box_size_without_header) {
-    return Error(heif_error_Invalid_input,
-                 heif_suberror_Invalid_box_size);
+    return {heif_error_Invalid_input,
+            heif_suberror_Invalid_box_size};
   }
 
 
@@ -740,6 +825,31 @@ std::vector<std::shared_ptr<Box>> Box::get_child_boxes(uint32_t short_type) cons
 }
 
 
+bool Box::operator==(const Box& other) const
+{
+  if (this->get_short_type() != other.get_short_type()) {
+    return false;
+  }
+
+  StreamWriter writer1;
+  StreamWriter writer2;
+
+  this->write(writer1);
+  other.write(writer2);
+
+  return writer1.get_data() == writer2.get_data();
+}
+
+
+bool Box::equal(const std::shared_ptr<Box>& box1, const std::shared_ptr<Box>& box2)
+{
+    if (!box1 || !box2) {
+        return false;
+    }
+    return *box1 == *box2;
+}
+
+
 Error Box::read_children(BitstreamRange& range, int max_number)
 {
   int count = 0;
@@ -824,6 +934,78 @@ void Box::derive_box_version_recursive()
 }
 
 
+Error Box_other::parse(BitstreamRange& range)
+{
+  if (has_fixed_box_size()) {
+    size_t len;
+    if (get_box_size() >= get_header_size()) {
+      auto len64 = get_box_size() - get_header_size();
+      if (len64 > MAX_BOX_SIZE) {
+        return {heif_error_Invalid_input,
+                heif_suberror_Security_limit_exceeded,
+                "Box size too large"};
+      }
+
+      len = static_cast<size_t>(len64);
+
+      m_data.resize(len);
+      range.read(m_data.data(), len);
+    }
+    else {
+      return {heif_error_Invalid_input,
+              heif_suberror_Invalid_box_size};
+    }
+  }
+  else {
+    // TODO: boxes until end of file (we will probably never need this)
+  }
+
+  return range.get_error();
+}
+
+
+Error Box_other::write(StreamWriter& writer) const
+{
+  size_t box_start = reserve_box_header_space(writer);
+
+  if (get_box_size() >= get_header_size()) {
+    writer.write(m_data);
+    prepend_header(writer, box_start);
+    return Error::Ok;
+  }
+  else {
+    return Error(heif_error_Invalid_input,
+                 heif_suberror_Invalid_box_size);
+  }
+}
+
+
+std::string Box_other::dump(Indent& indent) const
+{
+  std::ostringstream sstr;
+
+  sstr << BoxHeader::dump(indent);
+
+  // --- show raw box content
+
+  size_t len = 0;
+  if (get_box_size() >= get_header_size()) {
+    // We can cast because if it does not fit, it would fail during parsing.
+    len = static_cast<size_t>(get_box_size() - get_header_size());
+  }
+  else {
+    sstr << indent << "invalid box size " << get_box_size() << " (smaller than header)\n";
+    return sstr.str();
+  }
+
+  sstr << write_raw_data_as_hex(m_data.data(), len,
+                                "data: ",
+                                "      ");
+
+  return sstr.str();
+}
+
+
 Error Box_ftyp::parse(BitstreamRange& range)
 {
   m_major_brand = range.read32();
@@ -907,6 +1089,10 @@ Error Box_meta::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() != 0) {
+    return unsupported_version_error("meta");
+  }
+
   /*
   uint64_t boxSizeLimit;
   if (get_box_size() == BoxHeader::size_until_end_of_file) {
@@ -931,10 +1117,25 @@ std::string Box_meta::dump(Indent& indent) const
 }
 
 
+Error FullBox::unsupported_version_error(const char* box) const
+{
+  std::stringstream sstr;
+  sstr << box << " box data version " << ((int) m_version) << " is not implemented yet";
+
+  return {heif_error_Unsupported_feature,
+          heif_suberror_Unsupported_data_version,
+          sstr.str()};
+}
+
+
 Error Box_hdlr::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() != 0) {
+    return unsupported_version_error("hdlr");
+  }
+
   m_pre_defined = range.read32();
   m_handler_type = range.read32();
 
@@ -983,6 +1184,11 @@ Error Box_pitm::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() > 1) {
+    return unsupported_version_error("pitm");
+  }
+
+
   if (get_version() == 0) {
     m_item_ID = range.read16();
   }
@@ -1035,14 +1241,14 @@ Error Box_pitm::write(StreamWriter& writer) const
 
 Error Box_iloc::parse(BitstreamRange& range)
 {
-  /*
-  printf("box size: %d\n",get_box_size());
-  printf("header size: %d\n",get_header_size());
-  printf("start limit: %d\n",sizeLimit);
-  */
-
   parse_full_box_header(range);
 
+  if (get_version() > 2) {
+    return unsupported_version_error("iloc");
+  }
+
+  const int version = get_version();
+
   uint16_t values4 = range.read16();
 
   int offset_size = (values4 >> 12) & 0xF;
@@ -1050,15 +1256,15 @@ Error Box_iloc::parse(BitstreamRange& range)
   int base_offset_size = (values4 >> 4) & 0xF;
   int index_size = 0;
 
-  if (get_version() >= 1) {
+  if (version == 1 || version == 2) {
     index_size = (values4 & 0xF);
   }
 
-  uint32_t item_count;
-  if (get_version() < 2) {
+  uint32_t item_count = 0;
+  if (version < 2) {
     item_count = range.read16();
   }
-  else {
+  else if (version == 2) {
     item_count = range.read32();
   }
 
@@ -1076,14 +1282,14 @@ Error Box_iloc::parse(BitstreamRange& range)
   for (uint32_t i = 0; i < item_count; i++) {
     Item item;
 
-    if (get_version() < 2) {
+    if (version < 2) {
       item.item_ID = range.read16();
     }
-    else {
+    else if (version == 2) {
       item.item_ID = range.read32();
     }
 
-    if (get_version() >= 1) {
+    if (version >= 1) {
       values4 = range.read16();
       item.construction_method = (values4 & 0xF);
     }
@@ -1100,6 +1306,7 @@ Error Box_iloc::parse(BitstreamRange& range)
     }
 
     int extent_count = range.read16();
+
     // Sanity check.
     if (extent_count > MAX_ILOC_EXTENTS_PER_ITEM) {
       std::stringstream sstr;
@@ -1114,12 +1321,14 @@ Error Box_iloc::parse(BitstreamRange& range)
     for (int e = 0; e < extent_count; e++) {
       Extent extent;
 
-      if (index_size == 4) {
-        extent.index = range.read32();
-      }
-      else if (index_size == 8) {
-        extent.index = ((uint64_t) range.read32()) << 32;
-        extent.index |= range.read32();
+      if ((version == 1 || version == 2) && index_size > 0) {
+        if (index_size == 4) {
+          extent.index = range.read32();
+        }
+        else if (index_size == 8) {
+          extent.index = ((uint64_t) range.read32()) << 32;
+          extent.index |= range.read32();
+        }
       }
 
       extent.offset = 0;
@@ -1149,8 +1358,6 @@ Error Box_iloc::parse(BitstreamRange& range)
     }
   }
 
-  //printf("end limit: %d\n",sizeLimit);
-
   return range.get_error();
 }
 
@@ -1268,9 +1475,9 @@ Error Box_iloc::read_data(const Item& item,
     }
     else {
       std::stringstream sstr;
-      sstr << "Item construction method " << item.construction_method << " not implemented";
+      sstr << "Item construction method " << (int) item.construction_method << " not implemented";
       return Error(heif_error_Unsupported_feature,
-                   heif_suberror_No_idat_box,
+                   heif_suberror_Unsupported_item_construction_method,
                    sstr.str());
     }
   }
@@ -1550,10 +1757,30 @@ void Box_iloc::patch_iloc_header(StreamWriter& writer) const
 }
 
 
+/*
+ *                     version <= 1    version 2   version > 2    mime     uri
+ * -----------------------------------------------------------------------------------------------
+ * item id               16               16           32          16/32   16/32
+ * protection index      16               16           16          16      16
+ * item type             -                yes          yes         yes     yes
+ * item name             yes              yes          yes         yes     yes
+ * content type          yes              -            -           yes     -
+ * content encoding      yes              -            -           yes     -
+ * hidden item           -                yes          yes         yes     yes
+ * item uri type         -                -            -           -       yes
+ *
+ * Note: HEIF does not allow version 0 and version 1 boxes ! (see 23008-12, 10.2.1)
+ */
+
 Error Box_infe::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  // only versions 2,3 are required by HEIF
+  if (get_version() > 3) {
+    return unsupported_version_error("infe");
+  }
+
   if (get_version() <= 1) {
     m_item_ID = range.read16();
     m_item_protection_index = range.read16();
@@ -1680,11 +1907,18 @@ std::string Box_infe::dump(Indent& indent) const
   sstr << indent << "item_ID: " << m_item_ID << "\n"
        << indent << "item_protection_index: " << m_item_protection_index << "\n"
        << indent << "item_type: " << m_item_type << "\n"
-       << indent << "item_name: " << m_item_name << "\n"
-       << indent << "content_type: " << m_content_type << "\n"
-       << indent << "content_encoding: " << m_content_encoding << "\n"
-       << indent << "item uri type: " << m_item_uri_type << "\n"
-       << indent << "hidden item: " << std::boolalpha << m_hidden_item << "\n";
+       << indent << "item_name: " << m_item_name << "\n";
+
+  if (m_item_type == "mime") {
+    sstr << indent << "content_type: " << m_content_type << "\n"
+         << indent << "content_encoding: " << m_content_encoding << "\n";
+  }
+
+  if (m_item_type == "uri ") {
+    sstr << indent << "item uri type: " << m_item_uri_type << "\n";
+  }
+
+  sstr << indent << "hidden item: " << std::boolalpha << m_hidden_item << "\n";
 
   return sstr.str();
 }
@@ -1694,6 +1928,11 @@ Error Box_iinf::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  // TODO: there are several images in circulation that have an iinf version=2. We should not enforce this with a hard error.
+  if (false && get_version() > 1) {
+    return unsupported_version_error("iinf");
+  }
+
   int nEntries_size = (get_version() > 0) ? 4 : 2;
 
   uint32_t item_count;
@@ -1771,6 +2010,17 @@ std::string Box_iprp::dump(Indent& indent) const
 }
 
 
+int Box_ipco::find_or_append_child_box(const std::shared_ptr<Box>& box)
+{
+  for (int i = 0; i < (int) m_children.size(); i++) {
+    if (Box::equal(m_children[i], box)) {
+      return i;
+    }
+  }
+  return append_child_box(box);
+}
+
+
 Error Box_ipco::parse(BitstreamRange& range)
 {
   //parse_full_box_header(range);
@@ -1794,6 +2044,10 @@ Error Box_pixi::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() != 0) {
+    return unsupported_version_error("pixi");
+  }
+
   StreamReader::grow_status status;
   uint8_t num_channels = range.read8();
   status = range.wait_for_available_bytes(num_channels);
@@ -2101,6 +2355,10 @@ Error Box_ispe::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() != 0) {
+    return unsupported_version_error("ispe");
+  }
+
   m_image_width = range.read32();
   m_image_height = range.read32();
 
@@ -2133,10 +2391,28 @@ Error Box_ispe::write(StreamWriter& writer) const
 }
 
 
+bool Box_ispe::operator==(const Box& other) const
+{
+  const auto* other_ispe = dynamic_cast<const Box_ispe*>(&other);
+  if (other_ispe == nullptr) {
+    return false;
+  }
+
+  return (m_image_width == other_ispe->m_image_width &&
+          m_image_height == other_ispe->m_image_height);
+}
+
+
 Error Box_ipma::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  // TODO: is there any specification of allowed values for the ipma version in the HEIF standards?
+
+  if (get_version() > 1) {
+    return unsupported_version_error("ipma");
+  }
+
   uint32_t entry_cnt = range.read32();
   for (uint32_t i = 0; i < entry_cnt && !range.error() && !range.eof(); i++) {
     Entry entry;
@@ -2319,6 +2595,10 @@ Error Box_auxC::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() != 0) {
+    return unsupported_version_error("auxC");
+  }
+
   m_aux_type = range.read_string();
 
   while (!range.eof()) {
@@ -2528,6 +2808,21 @@ std::string Box_clap::dump(Indent& indent) const
 }
 
 
+double Box_clap::left(int image_width) const
+{
+  Fraction pcX = m_horizontal_offset + Fraction(image_width - 1, 2);
+  Fraction left = pcX - (m_clean_aperture_width - 1) / 2;
+  return left.to_double();
+}
+
+double Box_clap::top(int image_height) const
+{
+  Fraction pcY = m_vertical_offset + Fraction(image_height - 1, 2);
+  Fraction top = pcY - (m_clean_aperture_height - 1) / 2;
+  return top.to_double();
+}
+
+
 int Box_clap::left_rounded(int image_width) const
 {
   // pcX = horizOff + (width  - 1)/2
@@ -2591,6 +2886,10 @@ Error Box_iref::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() > 1) {
+    return unsupported_version_error("iref");
+  }
+
   while (!range.eof()) {
     Reference ref;
 
@@ -2925,34 +3224,7 @@ Error Box_grpl::parse(BitstreamRange& range)
 {
   //parse_full_box_header(range);
 
-  //return read_children(range);
-
-  while (!range.eof()) {
-    EntityGroup group;
-    Error err = group.header.parse_header(range);
-    if (err != Error::Ok) {
-      return err;
-    }
-
-    err = group.header.parse_full_box_header(range);
-    if (err != Error::Ok) {
-      return err;
-    }
-
-    group.group_id = range.read32();
-    uint32_t nEntities = range.read32();
-    for (uint32_t i = 0; i < nEntities; i++) {
-      if (range.eof()) {
-        break;
-      }
-
-      group.entity_ids.push_back(range.read32());
-    }
-
-    m_entity_groups.push_back(group);
-  }
-
-  return range.get_error();
+  return read_children(range); // should we pass the parsing context 'grpl' or are the box types unique?
 }
 
 
@@ -2960,23 +3232,131 @@ std::string Box_grpl::dump(Indent& indent) const
 {
   std::ostringstream sstr;
   sstr << Box::dump(indent);
+  sstr << dump_children(indent);
+  return sstr.str();
+}
 
-  for (const auto& group : m_entity_groups) {
-    sstr << indent << "group type: " << group.header.get_type_string() << "\n"
-         << indent << "| group id: " << group.group_id << "\n"
-         << indent << "| entity IDs: ";
 
-    for (uint32_t id : group.entity_ids) {
-      sstr << id << " ";
+Error Box_EntityToGroup::parse(BitstreamRange& range)
+{
+  Error err = parse_full_box_header(range);
+  if (err != Error::Ok) {
+    return err;
+  }
+
+  group_id = range.read32();
+  uint32_t nEntities = range.read32();
+  for (uint32_t i = 0; i < nEntities; i++) {
+    if (range.eof()) {
+      break;
     }
 
-    sstr << "\n";
+    entity_ids.push_back(range.read32());
+  }
+
+  return Error::Ok;
+}
+
+std::string Box_EntityToGroup::dump(Indent& indent) const
+{
+  std::ostringstream sstr;
+  sstr << Box::dump(indent);
+
+  sstr << indent << "group id: " << group_id << "\n"
+       << indent << "entity IDs: ";
+
+  bool first = true;
+  for (uint32_t id : entity_ids) {
+    if (first) {
+      first = false;
+    }
+    else {
+      sstr << ' ';
+    }
+
+    sstr << id;
+  }
+
+  sstr << "\n";
+
+  return sstr.str();
+}
+
+
+Error Box_ster::parse(BitstreamRange& range)
+{
+  Error err = Box_EntityToGroup::parse(range);
+  if (err) {
+    return err;
+  }
+
+  if (entity_ids.size() != 2) {
+    return {heif_error_Invalid_input,
+            heif_suberror_Invalid_box_size,
+            "'ster' entity group does not exists of exactly two images"};
+  }
+
+  return Error::Ok;
+}
+
+
+std::string Box_ster::dump(Indent& indent) const
+{
+  std::ostringstream sstr;
+  sstr << Box::dump(indent);
+
+  sstr << indent << "group id: " << group_id << "\n"
+       << indent << "left image ID: " << entity_ids[0] << "\n"
+       << indent << "right image ID: " << entity_ids[1] << "\n";
+
+  return sstr.str();
+}
+
+
+
+Error Box_pymd::parse(BitstreamRange& range)
+{
+  Error err = Box_EntityToGroup::parse(range);
+  if (err) {
+    return err;
+  }
+
+  tile_size_x = range.read16();
+  tile_size_y = range.read16();
+
+  for (size_t i = 0; i < entity_ids.size(); i++) {
+    LayerInfo layer{};
+    layer.layer_binning = range.read16();
+    layer.tiles_in_layer_row_minus1 = range.read16();
+    layer.tiles_in_layer_column_minus1 = range.read16();
+
+    m_layer_infos.push_back(layer);
+  }
+
+  return Error::Ok;
+}
+
+std::string Box_pymd::dump(Indent& indent) const
+{
+  std::ostringstream sstr;
+  sstr << Box_EntityToGroup::dump(indent);
+
+  sstr << indent << "tile size: " << tile_size_x << "x" << tile_size_y << "\n";
+
+  int layerNr = 0;
+  for (const auto& layer : m_layer_infos) {
+    sstr << indent << "layer " << layerNr << ":\n"
+         << indent << "| binning: " << layer.layer_binning << "\n"
+         << indent << "| tiles: " << (layer.tiles_in_layer_row_minus1 + 1) << "x" << (layer.tiles_in_layer_column_minus1 + 1) << "\n";
+
+    layerNr++;
   }
 
   return sstr.str();
 }
 
 
+
 Error Box_dinf::parse(BitstreamRange& range)
 {
   //parse_full_box_header(range);
@@ -2999,6 +3379,10 @@ Error Box_dref::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() != 0) {
+    return unsupported_version_error("dref");
+  }
+
   uint32_t nEntities = range.read32();
 
   /*
@@ -3042,6 +3426,10 @@ Error Box_url::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
 
+  if (get_version() > 0) {
+    return unsupported_version_error("url");
+  }
+
   m_location = range.read_string();
 
   return range.get_error();
@@ -3063,6 +3451,11 @@ std::string Box_url::dump(Indent& indent) const
 Error Box_udes::parse(BitstreamRange& range)
 {
   parse_full_box_header(range);
+
+  if (get_version() > 0) {
+    return unsupported_version_error("udes");
+  }
+
   m_lang = range.read_string();
   m_name = range.read_string();
   m_description = range.read_string();
@@ -3091,3 +3484,478 @@ Error Box_udes::write(StreamWriter& writer) const
   prepend_header(writer, box_start);
   return Error::Ok;
 }
+
+
+void Box_cmin::RelativeIntrinsicMatrix::compute_focal_length(int image_width, int image_height,
+                                                             double& out_focal_length_x, double& out_focal_length_y) const
+{
+  out_focal_length_x = focal_length_x * image_width;
+
+  if (is_anisotropic) {
+    out_focal_length_y = focal_length_y * image_height;
+  }
+  else {
+    out_focal_length_y = out_focal_length_x;
+  }
+}
+
+
+void Box_cmin::RelativeIntrinsicMatrix::compute_principal_point(int image_width, int image_height,
+                                                                double& out_principal_point_x, double& out_principal_point_y) const
+{
+  out_principal_point_x = principal_point_x * image_width;
+  out_principal_point_y = principal_point_y * image_height;
+}
+
+
+Box_cmin::AbsoluteIntrinsicMatrix Box_cmin::RelativeIntrinsicMatrix::to_absolute(int image_width, int image_height) const
+{
+  AbsoluteIntrinsicMatrix m{};
+  compute_focal_length(image_width, image_height, m.focal_length_x, m.focal_length_y);
+  compute_principal_point(image_width, image_height, m.principal_point_x, m.principal_point_y);
+  m.skew = skew;
+
+  return m;
+}
+
+
+std::string Box_cmin::dump(Indent& indent) const
+{
+  std::ostringstream sstr;
+  sstr << Box::dump(indent);
+  sstr << indent << "principal-point: " << m_matrix.principal_point_x << ", " << m_matrix.principal_point_y << "\n";
+  if (m_matrix.is_anisotropic) {
+    sstr << indent << "focal-length: " << m_matrix.focal_length_x << ", " << m_matrix.focal_length_y << "\n";
+    sstr << indent << "skew: " << m_matrix.skew << "\n";
+  }
+  else {
+    sstr << indent << "focal-length: " << m_matrix.focal_length_x << "\n";
+    sstr << indent << "no skew\n";
+  }
+
+  return sstr.str();
+}
+
+
+Error Box_cmin::parse(BitstreamRange& range)
+{
+  parse_full_box_header(range);
+
+  if (get_version() > 0) {
+    return unsupported_version_error("cmin");
+  }
+
+  m_denominatorShift = (get_flags() & 0x1F00) >> 8;
+  uint32_t denominator = (1U << m_denominatorShift);
+
+  m_matrix.focal_length_x = range.read32s() / (double)denominator;
+  m_matrix.principal_point_x = range.read32s() / (double)denominator;
+  m_matrix.principal_point_y = range.read32s() / (double)denominator;
+
+  if (get_flags() & 1) {
+    m_skewDenominatorShift = ((get_flags()) & 0x1F0000) >> 16;
+    uint32_t skewDenominator = (1U << m_skewDenominatorShift);
+
+    m_matrix.focal_length_y = range.read32s() / (double)denominator;
+    m_matrix.skew = range.read32s() / (double)skewDenominator;
+
+    m_matrix.is_anisotropic = true;
+  }
+  else {
+    m_matrix.is_anisotropic = false;
+    m_matrix.focal_length_y = 0;
+    m_matrix.skew = 0;
+  }
+  return range.get_error();
+}
+
+
+static uint32_t get_signed_fixed_point_shift(double v)
+{
+  if (v==0) {
+    return 31;
+  }
+
+  v = std::abs(v);
+
+  uint32_t shift = 0;
+  while (v < (1<<30)) {
+    v *= 2;
+    shift++;
+
+    if (shift==31) {
+      return shift;
+    }
+  }
+
+  return shift;
+}
+
+
+void Box_cmin::set_intrinsic_matrix(RelativeIntrinsicMatrix matrix)
+{
+  m_matrix = matrix;
+
+  uint32_t flags = 0;
+  flags |= matrix.is_anisotropic ? 1 : 0;
+
+  uint32_t shift_fx = get_signed_fixed_point_shift(matrix.focal_length_x);
+  uint32_t shift_px = get_signed_fixed_point_shift(matrix.principal_point_x);
+  uint32_t shift_py = get_signed_fixed_point_shift(matrix.principal_point_y);
+  m_denominatorShift = std::min(std::min(shift_fx, shift_px), shift_py);
+
+  if (matrix.is_anisotropic) {
+    uint32_t shift_fy = get_signed_fixed_point_shift(matrix.focal_length_y);
+    m_denominatorShift = std::min(m_denominatorShift, shift_fy);
+
+    m_skewDenominatorShift = get_signed_fixed_point_shift(matrix.skew);
+  }
+  else {
+    m_skewDenominatorShift = 0;
+  }
+
+  flags |= (m_denominatorShift << 8);
+  flags |= (m_skewDenominatorShift << 16);
+
+  set_flags(flags);
+}
+
+
+Error Box_cmin::write(StreamWriter& writer) const
+{
+  size_t box_start = reserve_box_header_space(writer);
+
+  uint32_t denominator = (1U << m_denominatorShift);
+
+  writer.write32s(static_cast<int32_t>(m_matrix.focal_length_x * denominator));
+  writer.write32s(static_cast<int32_t>(m_matrix.principal_point_x * denominator));
+  writer.write32s(static_cast<int32_t>(m_matrix.principal_point_y * denominator));
+
+  if (get_flags() & 1) {
+    writer.write32s(static_cast<int32_t>(m_matrix.focal_length_y * denominator));
+
+    uint32_t skewDenominator = (1U << m_skewDenominatorShift);
+    writer.write32s(static_cast<int32_t>(m_matrix.skew * skewDenominator));
+  }
+
+  prepend_header(writer, box_start);
+
+  return Error::Ok;
+}
+
+
+std::array<double,9> mul(const std::array<double,9>& a, const std::array<double,9>& b)
+{
+  std::array<double,9> m;
+
+  m[0] = a[0]*b[0] + a[1]*b[3] + a[2]*b[6];
+  m[1] = a[0]*b[1] + a[1]*b[4] + a[2]*b[7];
+  m[2] = a[0]*b[2] + a[1]*b[5] + a[2]*b[8];
+
+  m[3] = a[3]*b[0] + a[4]*b[3] + a[5]*b[6];
+  m[4] = a[3]*b[1] + a[4]*b[4] + a[5]*b[7];
+  m[5] = a[3]*b[2] + a[4]*b[5] + a[5]*b[8];
+
+  m[6] = a[6]*b[0] + a[7]*b[3] + a[8]*b[6];
+  m[7] = a[6]*b[1] + a[7]*b[4] + a[8]*b[7];
+  m[8] = a[6]*b[2] + a[7]*b[5] + a[8]*b[8];
+
+  return m;
+}
+
+
+std::array<double,9> Box_cmex::ExtrinsicMatrix::calculate_rotation_matrix() const
+{
+  std::array<double,9> m{};
+
+  if (rotation_as_quaternions) {
+    double qx = quaternion_x;
+    double qy = quaternion_y;
+    double qz = quaternion_z;
+    double qw = quaternion_w;
+
+    m[0] = 1-2*(qy*qy+qz*qz);
+    m[1] = 2*(qx*qy-qz*qw);
+    m[2] = 2*(qx*qz+qy*qw);
+    m[3] = 2*(qx*qy+qz*qw);
+    m[4] = 1-2*(qx*qx+qz*qz);
+    m[5] = 2*(qy*qz-qx*qw);
+    m[6] = 2*(qx*qz-qy*qw);
+    m[7] = 2*(qy*qz+qx*qw);
+    m[8] = 1-2*(qx*qx+qy*qy);
+  }
+  else {
+    // This rotation order fits the conformance data
+    // https://github.com/MPEGGroup/FileFormatConformance
+    // branch m62054_extrinsics : FileFormatConformance/data/file_features/under_consideration/ex_in_trinsics/extrinsic_rotation
+
+    std::array<double,9> m_yaw{};    // Z
+    std::array<double,9> m_pitch{};  // Y
+    std::array<double,9> m_roll{};   // X
+
+    const double d2r = M_PI/180;
+
+    double x = d2r * rotation_roll;
+    double y = d2r * rotation_pitch;
+    double z = d2r * rotation_yaw;
+
+    // X
+    m_roll[0] = 1;
+    m_roll[4] = m_roll[8] = cos(x);
+    m_roll[5] = -sin(x);
+    m_roll[7] = sin(x);
+
+    // Y
+    m_pitch[4] = 1;
+    m_pitch[0] = m_pitch[8] = cos(y);
+    m_pitch[6] = -sin(y);
+    m_pitch[2] = sin(y);
+
+    // Z
+    m_yaw[8] = 1;
+    m_yaw[0] = m_yaw[4] = cos(z);
+    m_yaw[1] = -sin(z);
+    m_yaw[3] = sin(z);
+
+    m = mul(m_yaw, mul(m_pitch, m_roll));
+  }
+
+  return m;
+}
+
+
+Error Box_cmex::parse(BitstreamRange& range)
+{
+  parse_full_box_header(range);
+
+  if (get_version() > 0) {
+    return unsupported_version_error("cmex");
+  }
+
+  m_matrix = ExtrinsicMatrix{};
+
+  if (get_flags() & pos_x_present) {
+    m_has_pos_x = true;
+    m_matrix.pos_x = range.read32s();
+  }
+
+  if (get_flags() & pos_y_present) {
+    m_has_pos_y = true;
+    m_matrix.pos_y = range.read32s();
+  }
+
+  if (get_flags() & pos_z_present) {
+    m_has_pos_z = true;
+    m_matrix.pos_z = range.read32s();
+  }
+
+  if (get_flags() & orientation_present) {
+    m_has_orientation = true;
+
+    if (get_version() == 0) {
+      bool use32bit = (get_flags() & rot_large_field_size);
+      int32_t quat_x = use32bit ? range.read32s() : range.read16s();
+      int32_t quat_y = use32bit ? range.read32s() : range.read16s();
+      int32_t quat_z = use32bit ? range.read32s() : range.read16s();
+
+      uint32_t div = 1U << (14 + (use32bit ? 16 : 0));
+
+      m_matrix.rotation_as_quaternions = true;
+      m_matrix.quaternion_x = quat_x / (double)div;
+      m_matrix.quaternion_y = quat_y / (double)div;
+      m_matrix.quaternion_z = quat_z / (double)div;
+
+      double q_sum = (m_matrix.quaternion_x * m_matrix.quaternion_x +
+                      m_matrix.quaternion_y * m_matrix.quaternion_y +
+                      m_matrix.quaternion_z * m_matrix.quaternion_z);
+
+      if (q_sum > 1.0) {
+        return Error(heif_error_Invalid_input,
+                     heif_suberror_Unspecified,
+                     "Invalid quaternion in extrinsic rotation matrix");
+      }
+
+      m_matrix.quaternion_w = sqrt(1 - q_sum);
+
+    } else if (get_version() == 1) {
+      uint32_t div = 1<<16;
+      m_matrix.rotation_yaw = range.read32s() / (double)div;
+      m_matrix.rotation_pitch = range.read32s() / (double)div;
+      m_matrix.rotation_roll = range.read32s() / (double)div;
+    }
+  }
+
+  if (get_flags() & id_present) {
+    m_has_world_coordinate_system_id = true;
+    m_matrix.world_coordinate_system_id = range.read32();
+  }
+
+  return range.get_error();
+}
+
+
+std::string Box_cmex::dump(Indent& indent) const
+{
+  std::ostringstream sstr;
+  sstr << Box::dump(indent);
+  sstr << indent << "camera position (um): ";
+  sstr << m_matrix.pos_x << " ; ";
+  sstr << m_matrix.pos_y << " ; ";
+  sstr << m_matrix.pos_z << "\n";
+
+  sstr << indent << "orientation ";
+  if (m_matrix.rotation_as_quaternions) {
+    sstr << "(quaterion)\n";
+    sstr << indent << "  q = ["
+         << m_matrix.quaternion_x << ";"
+         << m_matrix.quaternion_y << ";"
+         << m_matrix.quaternion_z << ";"
+         << m_matrix.quaternion_w << "]\n";
+  }
+  else {
+    sstr << "(angles)\n";
+    sstr << indent << "  yaw:   " << m_matrix.rotation_yaw << "\n";
+    sstr << indent << "  pitch: " << m_matrix.rotation_pitch << "\n";
+    sstr << indent << "  roll:  " << m_matrix.rotation_roll << "\n";
+  }
+
+  sstr << indent << "world coordinate system id: " << m_matrix.world_coordinate_system_id << "\n";
+
+  return sstr.str();
+}
+
+
+
+
+Error Box_cmex::set_extrinsic_matrix(ExtrinsicMatrix matrix)
+{
+  m_matrix = matrix;
+
+  uint32_t flags = 0;
+
+  m_has_pos_x = (matrix.pos_x != 0);
+  m_has_pos_y = (matrix.pos_y != 0);
+  m_has_pos_z = (matrix.pos_z != 0);
+
+  if (m_has_pos_x) {
+    flags |= pos_x_present;
+  }
+
+  if (m_has_pos_y) {
+    flags |= pos_y_present;
+  }
+
+  if (m_has_pos_z) {
+    flags |= pos_z_present;
+  }
+
+  if (matrix.rotation_as_quaternions) {
+    if (matrix.quaternion_x != 0 ||
+        matrix.quaternion_y != 0 ||
+        matrix.quaternion_z != 0) {
+      flags |= orientation_present;
+
+      double q_sum = (m_matrix.quaternion_x * m_matrix.quaternion_x +
+                      m_matrix.quaternion_y * m_matrix.quaternion_y +
+                      m_matrix.quaternion_z * m_matrix.quaternion_z);
+
+      if (q_sum > 1.0) {
+        return Error(heif_error_Invalid_input,
+                     heif_suberror_Unspecified,
+                     "Invalid quaternion in extrinsic rotation matrix");
+      }
+
+      if (matrix.quaternion_w < 0) {
+        matrix.quaternion_x = -matrix.quaternion_x;
+        matrix.quaternion_y = -matrix.quaternion_y;
+        matrix.quaternion_z = -matrix.quaternion_z;
+        matrix.quaternion_w = -matrix.quaternion_w;
+      }
+    }
+  }
+  else {
+    if (matrix.rotation_yaw != 0 ||
+        matrix.rotation_pitch != 0 ||
+        matrix.rotation_roll != 0) {
+      flags |= orientation_present;
+
+      if (matrix.rotation_yaw < -180.0 || matrix.rotation_yaw >= 180.0) {
+        return Error(heif_error_Invalid_input,
+                     heif_suberror_Unspecified,
+                     "Invalid yaw angle");
+      }
+
+      if (matrix.rotation_pitch < -90.0 || matrix.rotation_pitch > 90.0) {
+        return Error(heif_error_Invalid_input,
+                     heif_suberror_Unspecified,
+                     "Invalid pitch angle");
+      }
+
+      if (matrix.rotation_roll < -180.0 || matrix.rotation_roll >= 180.0) {
+        return Error(heif_error_Invalid_input,
+                     heif_suberror_Unspecified,
+                     "Invalid roll angle");
+      }
+    }
+  }
+
+  if (matrix.orientation_is_32bit) {
+    flags |= rot_large_field_size;
+  }
+
+  if (matrix.world_coordinate_system_id != 0) {
+    flags |= id_present;
+  }
+
+  set_flags(flags);
+  set_version(m_matrix.rotation_as_quaternions ? 0 : 1);
+
+  return Error::Ok;
+}
+
+
+Error Box_cmex::write(StreamWriter& writer) const
+{
+  size_t box_start = reserve_box_header_space(writer);
+
+  if (m_has_pos_x) {
+    writer.write32s(m_matrix.pos_x);
+  }
+
+  if (m_has_pos_y) {
+    writer.write32s(m_matrix.pos_y);
+  }
+
+  if (m_has_pos_z) {
+    writer.write32s(m_matrix.pos_z);
+  }
+
+  if (m_has_orientation) {
+    if (m_matrix.rotation_as_quaternions) {
+      if (m_matrix.orientation_is_32bit) {
+        writer.write32s(static_cast<int32_t>(m_matrix.quaternion_x * (1<<30)));
+        writer.write32s(static_cast<int32_t>(m_matrix.quaternion_y * (1<<30)));
+        writer.write32s(static_cast<int32_t>(m_matrix.quaternion_z * (1<<30)));
+      }
+      else {
+        writer.write16s(static_cast<int16_t>(m_matrix.quaternion_x * (1<<14)));
+        writer.write16s(static_cast<int16_t>(m_matrix.quaternion_y * (1<<14)));
+        writer.write16s(static_cast<int16_t>(m_matrix.quaternion_z * (1<<14)));
+      }
+    }
+    else {
+      writer.write32s(static_cast<int32_t>(m_matrix.rotation_yaw * (1<<16)));
+      writer.write32s(static_cast<int32_t>(m_matrix.rotation_pitch * (1<<16)));
+      writer.write32s(static_cast<int32_t>(m_matrix.rotation_roll * (1<<16)));
+    }
+  }
+
+  if (m_has_world_coordinate_system_id) {
+    writer.write32(m_matrix.world_coordinate_system_id);
+  }
+
+
+  prepend_header(writer, box_start);
+
+  return Error::Ok;
+}
