@@ -30,8 +30,8 @@
 
 #include "error.h"
 
-#include "heif.h"
-#include "heif_plugin.h"
+#include "libheif/heif.h"
+#include "libheif/heif_plugin.h"
 #include "bitstream.h"
 
 #include "box.h" // only for color_profile, TODO: maybe move the color_profiles to its own header
@@ -70,10 +70,11 @@ public:
 
   void set_max_decoding_threads(int max_threads) { m_max_decoding_threads = max_threads; }
 
+  // Sets the maximum size of both width and height of an image. The total limit
+  // of the image size (width * height) will be "maximum_size * maximum_size".
   void set_maximum_image_size_limit(int maximum_size)
   {
-    m_maximum_image_width_limit = maximum_size;
-    m_maximum_image_height_limit = maximum_size;
+    m_maximum_image_size_limit = int64_t(maximum_size) * maximum_size;
   }
 
   Error read(const std::shared_ptr<StreamReader>& reader);
@@ -97,6 +98,10 @@ public:
       m_aux_images.clear();
     }
 
+    Error check_resolution(uint32_t w, uint32_t h) const {
+      return m_heif_context->check_resolution(w, h);
+    }
+
     void set_resolution(int w, int h)
     {
       m_width = w;
@@ -134,10 +139,9 @@ public:
 
     // -- thumbnails
 
-    void set_is_thumbnail_of(heif_item_id id)
+    void set_is_thumbnail()
     {
       m_is_thumbnail = true;
-      m_thumbnail_ref_id = id;
     }
 
     void add_thumbnail(const std::shared_ptr<Image>& img) { m_thumbnails.push_back(img); }
@@ -149,11 +153,9 @@ public:
 
     // --- alpha channel
 
-    void set_is_alpha_channel_of(heif_item_id id, bool consumed)
+    void set_is_alpha_channel()
     {
       m_is_alpha_channel = true;
-      m_alpha_channel_ref_id = id;
-      m_implicitly_consumed_alpha = consumed;
     }
 
     void set_alpha_channel(std::shared_ptr<Image> img) { m_alpha_channel = std::move(img); }
@@ -169,10 +171,9 @@ public:
 
     // --- depth channel
 
-    void set_is_depth_channel_of(heif_item_id id)
+    void set_is_depth_channel()
     {
       m_is_depth_channel = true;
-      m_depth_channel_ref_id = id;
     }
 
     void set_depth_channel(std::shared_ptr<Image> img) { m_depth_channel = std::move(img); }
@@ -201,10 +202,9 @@ public:
 
     // --- generic aux image
 
-    void set_is_aux_image_of(heif_item_id id, const std::string& aux_type)
+    void set_is_aux_image(const std::string& aux_type)
     {
       m_is_aux_image = true;
-      m_aux_image_ref_id = id;
       m_aux_image_type = aux_type;
     }
 
@@ -222,8 +222,7 @@ public:
       else {
         std::vector<std::shared_ptr<Image>> auxImgs;
         for (const auto& aux : m_aux_images) {
-          if ((aux_image_filter & LIBHEIF_AUX_IMAGE_FILTER_OMIT_ALPHA) &&
-              aux->is_alpha_channel() && aux->m_implicitly_consumed_alpha) {
+          if ((aux_image_filter & LIBHEIF_AUX_IMAGE_FILTER_OMIT_ALPHA) && aux->is_alpha_channel()) {
             continue;
           }
 
@@ -278,6 +277,30 @@ public:
       }
     };
 
+    void set_intrinsic_matrix(const Box_cmin::RelativeIntrinsicMatrix& cmin) {
+      m_has_intrinsic_matrix = true;
+      m_intrinsic_matrix = cmin.to_absolute(get_ispe_width(), get_ispe_height());
+    }
+
+    bool has_intrinsic_matrix() const { return m_has_intrinsic_matrix; }
+
+    Box_cmin::AbsoluteIntrinsicMatrix& get_intrinsic_matrix() { return m_intrinsic_matrix; }
+
+    const Box_cmin::AbsoluteIntrinsicMatrix& get_intrinsic_matrix() const { return m_intrinsic_matrix; }
+
+
+    void set_extrinsic_matrix(const Box_cmex::ExtrinsicMatrix& cmex) {
+      m_has_extrinsic_matrix = true;
+      m_extrinsic_matrix = cmex;
+    }
+
+    bool has_extrinsic_matrix() const { return m_has_extrinsic_matrix; }
+
+    Box_cmex::ExtrinsicMatrix& get_extrinsic_matrix() { return m_extrinsic_matrix; }
+
+    const Box_cmex::ExtrinsicMatrix& get_extrinsic_matrix() const { return m_extrinsic_matrix; }
+
+
     void add_region_item_id(heif_item_id id) { m_region_item_ids.push_back(id); }
 
     const std::vector<heif_item_id>& get_region_item_ids() const { return m_region_item_ids; }
@@ -290,24 +313,19 @@ public:
     bool m_is_primary = false;
 
     bool m_is_thumbnail = false;
-    heif_item_id m_thumbnail_ref_id = 0;
 
     std::vector<std::shared_ptr<Image>> m_thumbnails;
 
     bool m_is_alpha_channel = false;
     bool m_premultiplied_alpha = false;
-    bool m_implicitly_consumed_alpha = false; // alpha data was integrated into main color image
-    heif_item_id m_alpha_channel_ref_id = 0;
     std::shared_ptr<Image> m_alpha_channel;
 
     bool m_is_depth_channel = false;
-    heif_item_id m_depth_channel_ref_id = 0;
     std::shared_ptr<Image> m_depth_channel;
 
     bool m_has_depth_representation_info = false;
     struct heif_depth_representation_info m_depth_representation_info;
 
-    heif_item_id m_aux_image_ref_id = 0;
     bool m_is_aux_image = false;
     std::string m_aux_image_type;
     std::vector<std::shared_ptr<Image>> m_aux_images;
@@ -320,9 +338,17 @@ public:
     bool m_miaf_compatible = true;
 
     std::vector<heif_item_id> m_region_item_ids;
+
+    bool m_has_intrinsic_matrix = false;
+    Box_cmin::AbsoluteIntrinsicMatrix m_intrinsic_matrix{};
+
+    bool m_has_extrinsic_matrix = false;
+    Box_cmex::ExtrinsicMatrix m_extrinsic_matrix{};
   };
 
-  std::shared_ptr<HeifFile> get_heif_file() { return m_heif_file; }
+  Error check_resolution(uint32_t width, uint32_t height) const;
+
+  std::shared_ptr<HeifFile> get_heif_file() const { return m_heif_file; }
 
   std::vector<std::shared_ptr<Image>> get_top_level_images() { return m_top_level_images; }
 
@@ -390,12 +416,25 @@ public:
                      enum heif_image_input_class input_class,
                      std::shared_ptr<Image>& out_image);
 
+  Error encode_grid(const std::vector<std::shared_ptr<HeifPixelImage>>& tiles,
+                    uint16_t rows,
+                    uint16_t columns,
+                    struct heif_encoder* encoder,
+                    const struct heif_encoding_options& options,
+                    std::shared_ptr<Image>& out_image);
+
   Error encode_image_as_hevc(const std::shared_ptr<HeifPixelImage>& image,
                              struct heif_encoder* encoder,
                              const struct heif_encoding_options& options,
                              enum heif_image_input_class input_class,
                              std::shared_ptr<Image>& out_image);
 
+  Error encode_image_as_vvc(const std::shared_ptr<HeifPixelImage>& image,
+                             struct heif_encoder* encoder,
+                             const struct heif_encoding_options& options,
+                             enum heif_image_input_class input_class,
+                             std::shared_ptr<Image>& out_image);
+
   Error encode_image_as_av1(const std::shared_ptr<HeifPixelImage>& image,
                             struct heif_encoder* encoder,
                             const struct heif_encoding_options& options,
@@ -449,8 +488,8 @@ public:
   Error add_XMP_metadata(const std::shared_ptr<Image>& master_image, const void* data, int size, heif_metadata_compression compression);
 
   Error add_generic_metadata(const std::shared_ptr<Image>& master_image, const void* data, int size,
-                             const char* item_type, const char* content_type,
-                             heif_metadata_compression compression);
+                             const char* item_type, const char* content_type, const char* item_uri_type,
+                             heif_metadata_compression compression, heif_item_id* out_item_id);
 
   heif_property_id add_property(heif_item_id targetItem, std::shared_ptr<Box> property, bool essential);
 
@@ -490,8 +529,8 @@ private:
 
   int m_max_decoding_threads = 4;
 
-  uint32_t m_maximum_image_width_limit;
-  uint32_t m_maximum_image_height_limit;
+  // Maximum image size in pixels (width * height).
+  uint64_t m_maximum_image_size_limit;
 
   std::vector<std::shared_ptr<RegionItem>> m_region_items;
 
@@ -506,7 +545,7 @@ private:
 
   Error decode_and_paste_tile_image(heif_item_id tileID,
                                     const std::shared_ptr<HeifPixelImage>& out_image,
-                                    int x0, int y0,
+                                    uint32_t x0, uint32_t y0,
                                     const heif_decoding_options& options) const;
 
   Error decode_derived_image(heif_item_id ID,
