@@ -102,6 +102,8 @@ const char* Error::get_error_string(heif_suberror_code err)
       return "No 'hdlr' box";
     case heif_suberror_No_hvcC_box:
       return "No 'hvcC' box";
+    case heif_suberror_No_vvcC_box:
+      return "No 'vvcC' box";
     case heif_suberror_No_av1C_box:
       return "No 'av1C' box";
     case heif_suberror_No_pitm_box:
@@ -158,12 +160,26 @@ const char* Error::get_error_string(heif_suberror_code err)
       return "Unknown NCLX matrix coefficients";
     case heif_suberror_Invalid_region_data:
       return "Invalid region item data";
+    case heif_suberror_No_ispe_property:
+      return "No ispe property";
+    case heif_suberror_Camera_intrinsic_matrix_undefined:
+      return "Camera intrinsic matrix undefined";
+    case heif_suberror_Camera_extrinsic_matrix_undefined:
+      return "Camera extrinsic matrix undefined";
+    case heif_suberror_Invalid_J2K_codestream:
+      return "Invalid JPEG 2000 codestream";
+    case heif_suberror_Decompression_invalid_data:
+      return "Invalid data in generic compression inflation";
+    case heif_suberror_No_icbr_box:
+      return "No 'icbr' box";
 
 
       // --- Memory_allocation_error ---
 
     case heif_suberror_Security_limit_exceeded:
       return "Security limit exceeded";
+    case heif_suberror_Compression_initialisation_error:
+      return "Compression initialisation method error";
 
       // --- Usage_error ---
 
@@ -200,6 +216,8 @@ const char* Error::get_error_string(heif_suberror_code err)
       return "Unsupported item construction method";
     case heif_suberror_Unsupported_header_compression_method:
       return "Unsupported header compression method";
+    case heif_suberror_Unsupported_generic_compression_method:
+      return "Unsupported generic compression method";
 
       // --- Encoder_plugin_error --
 
@@ -227,6 +245,8 @@ const char* Error::get_error_string(heif_suberror_code err)
       return "Trying to remove a plugin that is not loaded";
     case heif_suberror_Cannot_read_plugin_directory:
       return "Error while scanning the directory for plugins";
+    case heif_suberror_No_matching_decoder_installed:
+      return "No decoding plugin installed for this compression format";
   }
 
   assert(false);
