@@ -19,12 +19,14 @@
  */
 
 #include "file.h"
-#include "libheif/box.h"
+#include "box.h"
 #include "libheif/heif.h"
 #include "libheif/heif_properties.h"
-#include "libheif/jpeg2000.h"
-#include "libheif/jpeg.h"
-#include "libheif/vvc.h"
+#include "compression.h"
+#include "codecs/jpeg2000.h"
+#include "codecs/jpeg.h"
+#include "codecs/vvc.h"
+#include "codecs/uncompressed_box.h"
 
 #include <cstdint>
 #include <fstream>
@@ -44,11 +46,9 @@
 #include <windows.h>
 #endif
 
-#include "metadata_compression.h"
-#include "jpeg2000.h"
 
 #if WITH_UNCOMPRESSED_CODEC
-#include "uncompressed_image.h"
+#include "codecs/uncompressed_image.h"
 #endif
 
 // TODO: make this a decoder option
@@ -71,6 +71,28 @@ std::vector<heif_item_id> HeifFile::get_item_IDs() const
 }
 
 
+std::shared_ptr<const Box_infe> HeifFile::get_infe_box(heif_item_id ID) const
+{
+  auto iter = m_infe_boxes.find(ID);
+  if (iter == m_infe_boxes.end()) {
+    return nullptr;
+  }
+
+  return iter->second;
+}
+
+
+std::shared_ptr<Box_infe> HeifFile::get_infe_box(heif_item_id ID)
+{
+  auto iter = m_infe_boxes.find(ID);
+  if (iter == m_infe_boxes.end()) {
+    return nullptr;
+  }
+
+  return iter->second;
+}
+
+
 Error HeifFile::read_from_file(const char* input_filename)
 {
 #if defined(__MINGW32__) || defined(__MINGW64__) || defined(_MSC_VER)
@@ -185,6 +207,7 @@ void HeifFile::set_brand(heif_compression_format format, bool miaf_compatible)
       break;
 
     case heif_compression_JPEG2000:
+    case heif_compression_HTJ2K:
       m_ftyp_box->set_major_brand(fourcc("j2ki"));
       m_ftyp_box->set_minor_version(0);
       m_ftyp_box->add_compatible_brand(fourcc("mif1"));
@@ -397,7 +420,7 @@ Error HeifFile::parse_heif_file(BitstreamRange& range)
 
 
 Error HeifFile::check_for_ref_cycle(heif_item_id ID,
-                                    std::shared_ptr<Box_iref>& iref_box) const
+                                    const std::shared_ptr<Box_iref>& iref_box) const
 {
   std::unordered_set<heif_item_id> parent_items;
   return check_for_ref_cycle_recursion(ID, iref_box, parent_items);
@@ -405,7 +428,7 @@ Error HeifFile::check_for_ref_cycle(heif_item_id ID,
 
 
 Error HeifFile::check_for_ref_cycle_recursion(heif_item_id ID,
-                                    std::shared_ptr<Box_iref>& iref_box,
+                                    const std::shared_ptr<Box_iref>& iref_box,
                                     std::unordered_set<heif_item_id>& parent_items) const {
   if (parent_items.find(ID) != parent_items.end()) {
     return Error(heif_error_Invalid_input,
@@ -434,22 +457,16 @@ bool HeifFile::image_exists(heif_item_id ID) const
 }
 
 
-std::shared_ptr<Box_infe> HeifFile::get_infe(heif_item_id ID) const
+bool HeifFile::has_item_with_id(heif_item_id ID) const
 {
-  // --- get the image from the list of all images
-
-  auto image_iter = m_infe_boxes.find(ID);
-  if (image_iter == m_infe_boxes.end()) {
-    return nullptr;
-  }
-
-  return image_iter->second;
+  auto infe_box = get_infe_box(ID);
+  return infe_box != nullptr;
 }
 
 
 std::string HeifFile::get_item_type(heif_item_id ID) const
 {
-  auto infe_box = get_infe(ID);
+  auto infe_box = get_infe_box(ID);
   if (!infe_box) {
     return "";
   }
@@ -460,7 +477,7 @@ std::string HeifFile::get_item_type(heif_item_id ID) const
 
 std::string HeifFile::get_content_type(heif_item_id ID) const
 {
-  auto infe_box = get_infe(ID);
+  auto infe_box = get_infe_box(ID);
   if (!infe_box) {
     return "";
   }
@@ -470,7 +487,7 @@ std::string HeifFile::get_content_type(heif_item_id ID) const
 
 std::string HeifFile::get_item_uri_type(heif_item_id ID) const
 {
-  auto infe_box = get_infe(ID);
+  auto infe_box = get_infe_box(ID);
   if (!infe_box) {
     return "";
   }
@@ -506,6 +523,15 @@ heif_chroma HeifFile::get_image_chroma_from_configuration(heif_item_id imageID)
   }
 
 
+  // VVC
+
+  box = m_ipco_box->get_property_for_item_ID(imageID, m_ipma_box, fourcc("vvcC"));
+  std::shared_ptr<Box_vvcC> vvcC_box = std::dynamic_pointer_cast<Box_vvcC>(box);
+  if (vvcC_box) {
+    return (heif_chroma) (vvcC_box->get_configuration().chroma_format_idc);
+  }
+
+
   // AV1
 
   box = m_ipco_box->get_property_for_item_ID(imageID, m_ipma_box, fourcc("av1C"));
@@ -550,6 +576,23 @@ int HeifFile::get_luma_bits_per_pixel_from_configuration(heif_item_id imageID) c
   }
 
 
+  // VVC
+
+  if (image_type == "vvc1") {
+    auto box = m_ipco_box->get_property_for_item_ID(imageID, m_ipma_box, fourcc("vvcC"));
+    std::shared_ptr<Box_vvcC> vvcC_box = std::dynamic_pointer_cast<Box_vvcC>(box);
+    if (vvcC_box) {
+      Box_vvcC::configuration config = vvcC_box->get_configuration();
+      if (config.ptl_present_flag) {
+        return config.bit_depth_minus8 + 8;
+      }
+      else {
+        return 8; // TODO: what shall we do if the bit-depth is unknown? Use PIXI?
+      }
+    }
+  }
+
+  
   // AV1
 
   if (image_type == "av01") {
@@ -569,21 +612,6 @@ int HeifFile::get_luma_bits_per_pixel_from_configuration(heif_item_id imageID) c
     }
   }
 
-  // VVC
-
-  if (image_type == "vvc1") {
-    auto box = m_ipco_box->get_property_for_item_ID(imageID, m_ipma_box, fourcc("vvcC"));
-    std::shared_ptr<Box_vvcC> vvcC_box = std::dynamic_pointer_cast<Box_vvcC>(box);
-    if (vvcC_box) {
-      Box_vvcC::configuration config = vvcC_box->get_configuration();
-      if (config.bit_depth_present_flag) {
-        return config.bit_depth;
-      }
-      else {
-        return 8; // TODO: what shall we do if the bit-depth is unknown? Use PIXI?
-      }
-    }
-  }
 
   // JPEG
 
@@ -594,12 +622,12 @@ int HeifFile::get_luma_bits_per_pixel_from_configuration(heif_item_id imageID) c
   // JPEG 2000
 
   if (image_type == "j2k1") {
-    auto siz = jpeg2000_get_SIZ_segment(*this, imageID);
-    if (siz.components.empty()) {
+    JPEG2000MainHeader header;
+    Error err = header.parseHeader(*this, imageID);
+    if (err) {
       return -1;
     }
-
-    return siz.components[0].precision;
+    return header.get_precision(0);
   }
 
 #if WITH_UNCOMPRESSED_CODEC
@@ -629,6 +657,22 @@ int HeifFile::get_chroma_bits_per_pixel_from_configuration(heif_item_id imageID)
     }
   }
 
+  // VVC
+
+  if (image_type == "vvc1") {
+    auto box = m_ipco_box->get_property_for_item_ID(imageID, m_ipma_box, fourcc("vvcC"));
+    std::shared_ptr<Box_vvcC> vvcC_box = std::dynamic_pointer_cast<Box_vvcC>(box);
+    if (vvcC_box) {
+      Box_vvcC::configuration config = vvcC_box->get_configuration();
+      if (config.ptl_present_flag) {
+        return config.bit_depth_minus8 + 8;
+      }
+      else {
+        return 8; // TODO: what shall we do if the bit-depth is unknown? Use PIXI?
+      }
+    }
+  }
+
   // AV1
 
   if (image_type == "av01") {
@@ -657,13 +701,12 @@ int HeifFile::get_chroma_bits_per_pixel_from_configuration(heif_item_id imageID)
   // JPEG 2000
 
   if (image_type == "j2k1") {
-    auto siz = jpeg2000_get_SIZ_segment(*this, imageID);
-    if (siz.components.size() <= 1) {
+    JPEG2000MainHeader header;
+    Error err = header.parseHeader(*this, imageID);
+    if (err) {
       return -1;
     }
-
-    // TODO: this is a quick hack. It is more complicated for JPEG2000 because these can be any kind of colorspace (e.g. RGB).
-    return siz.components[1].precision;
+    return header.get_precision(1);
   }
 
   return -1;
@@ -709,7 +752,7 @@ Error HeifFile::get_compressed_image_data(heif_item_id ID, std::vector<uint8_t>*
                  heif_suberror_Nonexisting_item_referenced);
   }
 
-  auto infe_box = get_infe(ID);
+  auto infe_box = get_infe_box(ID);
   if (!infe_box) {
     return Error(heif_error_Usage_error,
                  heif_suberror_Nonexisting_item_referenced);
@@ -738,156 +781,86 @@ Error HeifFile::get_compressed_image_data(heif_item_id ID, std::vector<uint8_t>*
                  sstr.str());
   }
 
-  Error error = Error(heif_error_Unsupported_feature,
-                      heif_suberror_Unsupported_codec);
   if (item_type == "hvc1") {
     // --- --- --- HEVC
-
-    // --- get properties for this image
-
-    std::vector<std::shared_ptr<Box>> properties;
-    Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
-    if (err) {
-      return err;
-    }
-
-    // --- get codec configuration
-
-    std::shared_ptr<Box_hvcC> hvcC_box;
-    for (auto& prop : properties) {
-      if (prop->get_short_type() == fourcc("hvcC")) {
-        hvcC_box = std::dynamic_pointer_cast<Box_hvcC>(prop);
-        if (hvcC_box) {
-          break;
-        }
-      }
-    }
-
-    if (!hvcC_box) {
-      // Should always have an hvcC box, because we are checking this in
-      // heif_context::interpret_heif_file()
-      assert(false);
-      return Error(heif_error_Invalid_input,
-                   heif_suberror_No_hvcC_box);
-    }
-    else if (!hvcC_box->get_headers(data)) {
-      return Error(heif_error_Invalid_input,
-                   heif_suberror_No_item_data);
-    }
-
-    error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+    return get_compressed_image_data_hvc1(ID, data, item);
+  }
+  else if (item_type == "vvc1") {
+    // --- --- --- VVC
+    return get_compressed_image_data_vvc(ID, data, item);
   }
   else if (item_type == "av01") {
-    // --- --- --- AV1
-
-    // --- get properties for this image
-
-    std::vector<std::shared_ptr<Box>> properties;
-    Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
-    if (err) {
-      return err;
-    }
-
-    // --- get codec configuration
-
-    std::shared_ptr<Box_av1C> av1C_box;
-    for (auto& prop : properties) {
-      if (prop->get_short_type() == fourcc("av1C")) {
-        av1C_box = std::dynamic_pointer_cast<Box_av1C>(prop);
-        if (av1C_box) {
-          break;
-        }
-      }
-    }
-
-    if (!av1C_box) {
-      // Should always have an hvcC box, because we are checking this in
-      // heif_context::interpret_heif_file()
-      return Error(heif_error_Invalid_input,
-                   heif_suberror_No_av1C_box);
-    }
-    else if (!av1C_box->get_headers(data)) {
-      return Error(heif_error_Invalid_input,
-                   heif_suberror_No_item_data);
-    }
-
-    error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+    return get_compressed_image_data_av1(ID, data, item);
   }
   else if (item_type == "jpeg" ||
            (item_type == "mime" && get_content_type(ID) == "image/jpeg")) {
-
-    // --- check if 'jpgC' is present
-
-    std::vector<std::shared_ptr<Box>> properties;
-    Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
-    if (err) {
-      return err;
-    }
-
-    // --- get codec configuration
-
-    std::shared_ptr<Box_jpgC> jpgC_box;
-    for (auto& prop : properties) {
-      if (prop->get_short_type() == fourcc("jpgC")) {
-        jpgC_box = std::dynamic_pointer_cast<Box_jpgC>(prop);
-        if (jpgC_box) {
-          *data = jpgC_box->get_data();
-          break;
-        }
-      }
-    }
-
-    error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+    return get_compressed_image_data_jpeg(ID, data, item);
   }
   else if (item_type == "j2k1") {
-    std::vector<std::shared_ptr<Box>> properties;
-    Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
-    if (err) {
-      return err;
-    }
-
-    // --- get codec configuration
-
-    std::shared_ptr<Box_j2kH> j2kH_box;
-    for (auto& prop : properties) {
-      if (prop->get_short_type() == fourcc("j2kH")) {
-        j2kH_box = std::dynamic_pointer_cast<Box_j2kH>(prop);
-        if (j2kH_box) {
-          break;
-        }
-      }
-    }
-
-    if (!j2kH_box) {
-      // Should always have an j2kH box, because we are checking this in
-      // heif_context::interpret_heif_file()
-
-      //TODO - Correctly Find the j2kH box
-      // return Error(heif_error_Invalid_input,
-      //              heif_suberror_Unspecified);
-    }
-    // else if (!j2kH_box->get_headers(data)) {
-    //   return Error(heif_error_Invalid_input,
-    //                heif_suberror_No_item_data);
-    // }
-
-    error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+      return get_compressed_image_data_jpeg2000(ID, item, data);
   }
+#if WITH_UNCOMPRESSED_CODEC
+  else if (item_type == "unci") {
+    return get_compressed_image_data_uncompressed(ID, data, item);
+  }
+#endif
   else if (true ||  // fallback case for all kinds of generic metadata (e.g. 'iptc')
            item_type == "grid" ||
            item_type == "iovl" ||
            item_type == "Exif" ||
            (item_type == "mime" && content_type == "application/rdf+xml")) {
-
+    Error error;
     bool read_uncompressed = true;
     if (item_type == "mime") {
       std::string encoding = infe_box->get_content_encoding();
-      if (encoding == "deflate") {
-#if WITH_DEFLATE_HEADER_COMPRESSION
+      if (encoding == "compress_zlib") {
+#if HAVE_ZLIB
         read_uncompressed = false;
         std::vector<uint8_t> compressed_data;
         error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, &compressed_data);
-        *data = inflate(compressed_data);
+        if (error) {
+          return error;
+        }
+        error = decompress_zlib(compressed_data, data);
+        if (error) {
+          return error;
+        }
+#else
+        return Error(heif_error_Unsupported_feature,
+                     heif_suberror_Unsupported_header_compression_method,
+                     encoding);
+#endif
+      }
+      else if (encoding == "deflate") {
+#if HAVE_ZLIB
+        read_uncompressed = false;
+        std::vector<uint8_t> compressed_data;
+        error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, &compressed_data);
+        if (error) {
+          return error;
+        }
+        error = decompress_deflate(compressed_data, data);
+        if (error) {
+          return error;
+        }
+#else
+        return Error(heif_error_Unsupported_feature,
+                     heif_suberror_Unsupported_header_compression_method,
+                     encoding);
+#endif
+      }
+      else if (encoding == "br") {
+#if HAVE_BROTLI
+        read_uncompressed = false;
+        std::vector<uint8_t> compressed_data;
+        error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, &compressed_data);
+        if (error) {
+          return error;
+        }
+        error = decompress_brotli(compressed_data, data);
+        if (error) {
+          return error;
+        }
 #else
         return Error(heif_error_Unsupported_feature,
                      heif_suberror_Unsupported_header_compression_method,
@@ -897,15 +870,426 @@ Error HeifFile::get_compressed_image_data(heif_item_id ID, std::vector<uint8_t>*
     }
 
     if (read_uncompressed) {
-      error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+      return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+    }
+  }
+  return Error(heif_error_Unsupported_feature, heif_suberror_Unsupported_codec);
+}
+
+#if WITH_UNCOMPRESSED_CODEC
+// generic compression and uncompressed, per 23001-17
+const Error HeifFile::get_compressed_image_data_uncompressed(heif_item_id ID, std::vector<uint8_t> *data, const Box_iloc::Item *item) const
+{
+  std::vector<std::shared_ptr<Box>> properties;
+  Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
+  if (err) {
+    return err;
+  }
+
+  // --- get codec configuration
+
+  std::shared_ptr<Box_cmpC> cmpC_box;
+  std::shared_ptr<Box_icbr> icbr_box;
+  for (auto& prop : properties) {
+    if (prop->get_short_type() == fourcc("cmpC")) {
+      cmpC_box = std::dynamic_pointer_cast<Box_cmpC>(prop);
+    }
+    if (prop->get_short_type() == fourcc("icbr")) {
+      icbr_box = std::dynamic_pointer_cast<Box_icbr>(prop);
+    }
+    if (cmpC_box && icbr_box) {
+      break;
+    }
+  }
+  if (!cmpC_box) {
+    // assume no generic compression
+    return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+  }
+  if (!cmpC_box->get_must_decompress_individual_entities()) {
+    std::vector<uint8_t> compressed_data;
+    m_iloc_box->read_data(*item, m_input_stream, m_idat_box, &compressed_data);
+    return do_decompress_data(cmpC_box, compressed_data, data);
+  } else {
+    if (!icbr_box) {
+      std::stringstream sstr;
+      sstr << "cannot decode unci item requiring entity decompression without icbr box" << std::endl;
+      return Error(heif_error_Invalid_input,
+                  heif_suberror_No_icbr_box,
+                  sstr.str());
+    }
+    if (item->construction_method == 0) {
+      for (Box_icbr::ByteRange range: icbr_box->get_ranges()) {
+        // TODO: check errors
+        bool success = m_input_stream->seek(range.range_offset);
+        if (!success) {
+          return Error{heif_error_Invalid_input, heif_suberror_End_of_data, "error while seeking to generically compressed data"};
+        }
+        std::vector<uint8_t> compressed_range_bytes(range.range_size);
+        success = m_input_stream->read((char*) compressed_range_bytes.data(), static_cast<size_t>(compressed_range_bytes.size()));
+        if (!success) {
+          return Error{heif_error_Invalid_input, heif_suberror_End_of_data, "error while reading generically compressed data"};
+        }
+        std::vector<uint8_t> uncompressed_range_data;
+        Error err = do_decompress_data(cmpC_box, compressed_range_bytes, &uncompressed_range_data);
+        if (err) {
+          return err;
+        }
+        data->insert(data->end(), uncompressed_range_data.data(), uncompressed_range_data.data() + uncompressed_range_data.size());
+      }
+      return Error::Ok;
+    } else {
+      // TODO: implement...
+      std::stringstream sstr;
+      sstr << "cannot decode unci item from idat yet" << std::endl;
+      return Error(heif_error_Unsupported_feature,
+                  heif_suberror_Unsupported_data_version,
+                  sstr.str());
+    }
+  }
+  return Error(heif_error_Unsupported_feature, heif_suberror_Unsupported_codec);
+}
+
+const Error HeifFile::do_decompress_data(std::shared_ptr<Box_cmpC> &cmpC_box, std::vector<uint8_t> compressed_data, std::vector<uint8_t> *data) const
+{
+  if (cmpC_box->get_compression_type() == fourcc("brot")) {
+#if HAVE_BROTLI
+    return decompress_brotli(compressed_data, data);
+#else
+    std::stringstream sstr;
+    sstr << "cannot decode unci item with brotli compression - not enabled" << std::endl;
+    return Error(heif_error_Unsupported_feature,
+                 heif_suberror_Unsupported_generic_compression_method,
+                 sstr.str());
+#endif
+  } else if (cmpC_box->get_compression_type() == fourcc("zlib")) {
+#if HAVE_ZLIB
+    return decompress_zlib(compressed_data, data);
+#else
+    std::stringstream sstr;
+    sstr << "cannot decode unci item with zlib compression - not enabled" << std::endl;
+    return Error(heif_error_Unsupported_feature,
+                 heif_suberror_Unsupported_generic_compression_method,
+                 sstr.str());
+#endif
+  } else if (cmpC_box->get_compression_type() == fourcc("defl")) {
+#if HAVE_ZLIB
+    return decompress_deflate(compressed_data, data);
+#else
+    std::stringstream sstr;
+    sstr << "cannot decode unci item with deflate compression - not enabled" << std::endl;
+    return Error(heif_error_Unsupported_feature,
+                 heif_suberror_Unsupported_generic_compression_method,
+                 sstr.str());
+#endif
+  } else {
+    std::stringstream sstr;
+    sstr << "cannot decode unci item with unsupported compression type: " << cmpC_box->get_compression_type() << std::endl;
+    return Error(heif_error_Unsupported_feature,
+                 heif_suberror_Unsupported_generic_compression_method,
+                 sstr.str());
+  }
+}
+#endif
+
+const Error HeifFile::get_compressed_image_data_hvc1(heif_item_id ID, std::vector<uint8_t> *data, const Box_iloc::Item *item) const
+{
+  // --- get properties for this image
+  std::vector<std::shared_ptr<Box>> properties;
+  Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
+  if (err)
+  {
+    return err;
+  }
+
+  // --- get codec configuration
+
+  std::shared_ptr<Box_hvcC> hvcC_box;
+  for (auto &prop : properties)
+  {
+    if (prop->get_short_type() == fourcc("hvcC"))
+    {
+      hvcC_box = std::dynamic_pointer_cast<Box_hvcC>(prop);
+      if (hvcC_box)
+      {
+        break;
+      }
     }
   }
 
-  if (error != Error::Ok) {
+  if (!hvcC_box)
+  {
+    // Should always have an hvcC box, because we are checking this in
+    // heif_context::interpret_heif_file()
+    assert(false);
+    return Error(heif_error_Invalid_input,
+                 heif_suberror_No_hvcC_box);
+  }
+  else if (!hvcC_box->get_headers(data))
+  {
+    return Error(heif_error_Invalid_input,
+                 heif_suberror_No_item_data);
+  }
+
+  return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+}
+
+
+const Error HeifFile::get_compressed_image_data_vvc(heif_item_id ID, std::vector<uint8_t> *data, const Box_iloc::Item *item) const
+{
+
+  // --- get properties for this image
+
+  std::vector<std::shared_ptr<Box>> properties;
+  Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
+  if (err)
+  {
+    return err;
+  }
+
+  // --- get codec configuration
+
+  std::shared_ptr<Box_vvcC> vvcC_box;
+  for (auto &prop : properties)
+  {
+    if (prop->get_short_type() == fourcc("vvcC"))
+    {
+      vvcC_box = std::dynamic_pointer_cast<Box_vvcC>(prop);
+      if (vvcC_box)
+      {
+        break;
+      }
+    }
+  }
+
+  if (!vvcC_box)
+  {
+    assert(false);
+    return Error(heif_error_Invalid_input,
+                 heif_suberror_No_vvcC_box);
+  }
+  else if (!vvcC_box->get_headers(data))
+  {
+    return Error(heif_error_Invalid_input,
+                 heif_suberror_No_item_data);
+  }
+
+  return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+}
+
+const Error HeifFile::get_compressed_image_data_av1(heif_item_id ID, std::vector<uint8_t> *data, const Box_iloc::Item *item) const
+{
+  // --- --- --- AV1
+
+  // --- get properties for this image
+
+  std::vector<std::shared_ptr<Box>> properties;
+  Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
+  if (err)
+  {
+    return err;
+  }
+
+  // --- get codec configuration
+
+  std::shared_ptr<Box_av1C> av1C_box;
+  for (auto &prop : properties)
+  {
+    if (prop->get_short_type() == fourcc("av1C"))
+    {
+      av1C_box = std::dynamic_pointer_cast<Box_av1C>(prop);
+      if (av1C_box)
+      {
+        break;
+      }
+    }
+  }
+
+  if (!av1C_box)
+  {
+    return Error(heif_error_Invalid_input,
+                 heif_suberror_No_av1C_box);
+  }
+  else if (!av1C_box->get_headers(data))
+  {
+    return Error(heif_error_Invalid_input,
+                 heif_suberror_No_item_data);
+  }
+
+  return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+}
+
+const Error HeifFile::get_compressed_image_data_jpeg2000(heif_item_id ID, const Box_iloc::Item *item, std::vector<uint8_t> *data) const
+{
+  std::vector<std::shared_ptr<Box>> properties;
+  Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
+  if (err)
+  {
+    return err;
+  }
+
+  // --- get codec configuration
+
+  std::shared_ptr<Box_j2kH> j2kH_box;
+  for (auto &prop : properties)
+  {
+    if (prop->get_short_type() == fourcc("j2kH"))
+    {
+      j2kH_box = std::dynamic_pointer_cast<Box_j2kH>(prop);
+      if (j2kH_box)
+      {
+        break;
+      }
+    }
+  }
+
+  if (!j2kH_box)
+  {
+    // TODO - Correctly Find the j2kH box
+    //  return Error(heif_error_Invalid_input,
+    //               heif_suberror_Unspecified);
+  }
+  // else if (!j2kH_box->get_headers(data)) {
+  //   return Error(heif_error_Invalid_input,
+  //                heif_suberror_No_item_data);
+  // }
+
+  return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+}
+
+const Error HeifFile::get_compressed_image_data_jpeg(heif_item_id ID, std::vector<uint8_t> * data, const Box_iloc::Item *item) const
+{
+  // --- check if 'jpgC' is present
+  std::vector<std::shared_ptr<Box>> properties;
+  Error err = m_ipco_box->get_properties_for_item_ID(ID, m_ipma_box, properties);
+  if (err)
+  {
+    return err;
+  }
+
+  // --- get codec configuration
+
+  std::shared_ptr<Box_jpgC> jpgC_box;
+  for (auto &prop : properties)
+  {
+    if (prop->get_short_type() == fourcc("jpgC"))
+    {
+      jpgC_box = std::dynamic_pointer_cast<Box_jpgC>(prop);
+      if (jpgC_box)
+      {
+        *data = jpgC_box->get_data();
+        break;
+      }
+    }
+  }
+
+  return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, data);
+}
+
+
+Error HeifFile::get_item_data(heif_item_id ID, std::vector<uint8_t>* out_data, heif_metadata_compression* out_compression) const
+{
+  Error error;
+
+  auto infe_box = get_infe_box(ID);
+  if (!infe_box) {
+    return {heif_error_Usage_error,
+            heif_suberror_Nonexisting_item_referenced};
+  }
+
+  std::string item_type = infe_box->get_item_type();
+  std::string content_type = infe_box->get_content_type();
+
+  // --- get item
+
+  auto items = m_iloc_box->get_items();
+  const Box_iloc::Item* item = nullptr;
+  for (const auto& i : items) {
+    if (i.item_ID == ID) {
+      item = &i;
+      break;
+    }
+  }
+  if (!item) {
+    std::stringstream sstr;
+    sstr << "Item with ID " << ID << " has no data";
+
+    return {heif_error_Invalid_input,
+            heif_suberror_No_item_data,
+            sstr.str()};
+  }
+
+  // --- non 'mime' data (uncompressed)
+
+  if (item_type != "mime") {
+    if (out_compression) {
+      *out_compression = heif_metadata_compression_off;
+    }
+
+    return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, out_data);
+  }
+
+
+  // --- mime data
+
+  std::string encoding = infe_box->get_content_encoding();
+
+  heif_metadata_compression compression;
+
+  if (encoding.empty()) {
+    // shortcut for case of uncompressed mime data
+
+    if (out_compression) {
+      *out_compression = heif_metadata_compression_off;
+    }
+
+    return m_iloc_box->read_data(*item, m_input_stream, m_idat_box, out_data);
+  }
+  else if (encoding == "compress_zlib") {
+    compression = heif_metadata_compression_zlib;
+  }
+  else if (encoding == "deflate") {
+    compression = heif_metadata_compression_deflate;
+  }
+  else if (encoding == "br") {
+    compression = heif_metadata_compression_brotli;
+  }
+  else {
+    compression = heif_metadata_compression_unknown;
+  }
+
+  // read compressed data
+
+  std::vector<uint8_t> compressed_data;
+  error = m_iloc_box->read_data(*item, m_input_stream, m_idat_box, &compressed_data);
+  if (error) {
     return error;
   }
 
-  return Error::Ok;
+  // return compressed data, if we do not want to have it uncompressed
+
+  const bool do_decode = (out_compression == nullptr);
+  if (!do_decode) {
+    *out_compression = compression;
+    *out_data = std::move(compressed_data);
+    return Error::Ok;
+  }
+
+  // decompress the data
+
+  switch (compression) {
+#if HAVE_ZLIB
+    case heif_metadata_compression_zlib:
+      return decompress_zlib(compressed_data, out_data);
+    case heif_metadata_compression_deflate:
+      return decompress_deflate(compressed_data, out_data);
+#endif
+#if HAVE_BROTLI
+    case heif_metadata_compression_brotli:
+      return decompress_brotli(compressed_data, out_data);
+#endif
+    default:
+      return {heif_error_Unsupported_filetype, heif_suberror_Unsupported_header_compression_method};
+  }
 }
 
 
@@ -969,7 +1353,7 @@ void HeifFile::add_ispe_property(heif_item_id id, uint32_t width, uint32_t heigh
   auto ispe = std::make_shared<Box_ispe>();
   ispe->set_size(width, height);
 
-  int index = m_ipco_box->append_child_box(ispe);
+  int index = m_ipco_box->find_or_append_child_box(ispe);
 
   m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{false, uint16_t(index + 1)});
 }
@@ -980,15 +1364,15 @@ void HeifFile::add_clap_property(heif_item_id id, uint32_t clap_width, uint32_t
   auto clap = std::make_shared<Box_clap>();
   clap->set(clap_width, clap_height, image_width, image_height);
 
-  int index = m_ipco_box->append_child_box(clap);
+  int index = m_ipco_box->find_or_append_child_box(clap);
 
   m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{true, uint16_t(index + 1)});
 }
 
 
-heif_property_id HeifFile::add_property(heif_item_id id, std::shared_ptr<Box> property, bool essential)
+heif_property_id HeifFile::add_property(heif_item_id id, const std::shared_ptr<Box>& property, bool essential)
 {
-  int index = m_ipco_box->append_child_box(property);
+  int index = m_ipco_box->find_or_append_child_box(property);
 
   m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{essential, uint16_t(index + 1)});
 
@@ -1042,7 +1426,7 @@ void HeifFile::add_orientation_properties(heif_item_id id, heif_orientation orie
     auto irot = std::make_shared<Box_irot>();
     irot->set_rotation_ccw(rotation_ccw);
 
-    int index = m_ipco_box->append_child_box(irot);
+    int index = m_ipco_box->find_or_append_child_box(irot);
 
     m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{false, uint16_t(index + 1)});
   }
@@ -1051,7 +1435,7 @@ void HeifFile::add_orientation_properties(heif_item_id id, heif_orientation orie
     auto imir = std::make_shared<Box_imir>();
     imir->set_mirror_direction(mirror);
 
-    int index = m_ipco_box->append_child_box(imir);
+    int index = m_ipco_box->find_or_append_child_box(imir);
 
     m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{false, uint16_t(index + 1)});
   }
@@ -1067,12 +1451,77 @@ void HeifFile::add_pixi_property(heif_item_id id, uint8_t c1, uint8_t c2, uint8_
     pixi->add_channel_bits(c3);
   }
 
-  int index = m_ipco_box->append_child_box(pixi);
+  int index = m_ipco_box->find_or_append_child_box(pixi);
 
   m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{false, uint16_t(index + 1)});
 }
 
 
+void HeifFile::add_vvcC_property(heif_item_id id)
+{
+  auto vvcC = std::make_shared<Box_vvcC>();
+  int index = m_ipco_box->append_child_box(vvcC);
+
+  m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{true, uint16_t(index + 1)});
+}
+
+
+Error HeifFile::append_vvcC_nal_data(heif_item_id id, const std::vector<uint8_t>& nal_data)
+{
+  auto vvcC = std::dynamic_pointer_cast<Box_vvcC>(m_ipco_box->get_property_for_item_ID(id,
+                                                                                       m_ipma_box,
+                                                                                       fourcc("vvcC")));
+
+  if (vvcC) {
+    vvcC->append_nal_data(nal_data);
+    return Error::Ok;
+  }
+  else {
+    // Should always have an vvcC box, because we are checking this in
+    // heif_context::interpret_heif_file()
+    assert(false);
+    return Error(heif_error_Usage_error,
+                 heif_suberror_No_vvcC_box);
+  }
+}
+
+
+Error HeifFile::set_vvcC_configuration(heif_item_id id, const Box_vvcC::configuration& config)
+{
+  auto vvcC = std::dynamic_pointer_cast<Box_vvcC>(m_ipco_box->get_property_for_item_ID(id,
+                                                                                       m_ipma_box,
+                                                                                       fourcc("vvcC")));
+
+  if (vvcC) {
+    vvcC->set_configuration(config);
+    return Error::Ok;
+  }
+  else {
+    return Error(heif_error_Usage_error,
+                 heif_suberror_No_vvcC_box);
+  }
+}
+
+
+Error HeifFile::append_vvcC_nal_data(heif_item_id id, const uint8_t* data, size_t size)
+{
+  std::vector<std::shared_ptr<Box>> properties;
+
+  auto vvcC = std::dynamic_pointer_cast<Box_vvcC>(m_ipco_box->get_property_for_item_ID(id,
+                                                                                       m_ipma_box,
+                                                                                       fourcc("vvcC")));
+
+  if (vvcC) {
+    vvcC->append_nal_data(data, size);
+    return Error::Ok;
+  }
+  else {
+    return Error(heif_error_Usage_error,
+                 heif_suberror_No_vvcC_box);
+  }
+}
+
+
 void HeifFile::add_hvcC_property(heif_item_id id)
 {
   auto hvcC = std::make_shared<Box_hvcC>();
@@ -1138,35 +1587,19 @@ Error HeifFile::append_hvcC_nal_data(heif_item_id id, const uint8_t* data, size_
 }
 
 
-void HeifFile::add_av1C_property(heif_item_id id)
+void HeifFile::add_av1C_property(heif_item_id id, const Box_av1C::configuration& config)
 {
   auto av1C = std::make_shared<Box_av1C>();
-  int index = m_ipco_box->append_child_box(av1C);
+  av1C->set_configuration(config);
 
-  m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{true, uint16_t(index + 1)});
+  add_property(id, av1C, true);
 }
 
 
-Error HeifFile::set_av1C_configuration(heif_item_id id, const Box_av1C::configuration& config)
-{
-  auto av1C = std::dynamic_pointer_cast<Box_av1C>(m_ipco_box->get_property_for_item_ID(id,
-                                                                                       m_ipma_box,
-                                                                                       fourcc("av1C")));
-
-  if (av1C) {
-    av1C->set_configuration(config);
-    return Error::Ok;
-  }
-  else {
-    return Error(heif_error_Usage_error,
-                 heif_suberror_No_av1C_box);
-  }
-}
-
-std::shared_ptr<Box_j2kH> HeifFile::add_j2kH_property(heif_item_id id) 
+std::shared_ptr<Box_j2kH> HeifFile::add_j2kH_property(heif_item_id id)
 {
   auto j2kH = std::make_shared<Box_j2kH>();
-  int index = m_ipco_box->append_child_box(j2kH);
+  int index = m_ipco_box->append_child_box(j2kH); // do not deduplicate because this can have a child box
 
   m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{true, uint16_t(index + 1)});
 
@@ -1174,6 +1607,154 @@ std::shared_ptr<Box_j2kH> HeifFile::add_j2kH_property(heif_item_id id)
 }
 
 
+Result<heif_item_id> HeifFile::add_infe(const char* item_type, const uint8_t* data, size_t size)
+{
+  Result<heif_item_id> result;
+
+  // create an infe box describing what kind of data we are storing (this also creates a new ID)
+
+  auto infe_box = add_new_infe_box(item_type);
+  infe_box->set_hidden_item(true);
+
+  heif_item_id metadata_id = infe_box->get_item_ID();
+  result.value = metadata_id;
+
+  set_item_data(infe_box, data, size, heif_metadata_compression_off);
+
+  return result;
+}
+
+
+Result<heif_item_id> HeifFile::add_infe_mime(const char* content_type, heif_metadata_compression content_encoding, const uint8_t* data, size_t size)
+{
+  Result<heif_item_id> result;
+
+  // create an infe box describing what kind of data we are storing (this also creates a new ID)
+
+  auto infe_box = add_new_infe_box("mime");
+  infe_box->set_hidden_item(true);
+  infe_box->set_content_type(content_type);
+
+  heif_item_id metadata_id = infe_box->get_item_ID();
+  result.value = metadata_id;
+
+  set_item_data(infe_box, data, size, content_encoding);
+
+  return result;
+}
+
+
+Result<heif_item_id> HeifFile::add_precompressed_infe_mime(const char* content_type, std::string content_encoding, const uint8_t* data, size_t size)
+{
+  Result<heif_item_id> result;
+
+  // create an infe box describing what kind of data we are storing (this also creates a new ID)
+
+  auto infe_box = add_new_infe_box("mime");
+  infe_box->set_hidden_item(true);
+  infe_box->set_content_type(content_type);
+
+  heif_item_id metadata_id = infe_box->get_item_ID();
+  result.value = metadata_id;
+
+  set_precompressed_item_data(infe_box, data, size, content_encoding);
+
+  return result;
+}
+
+
+Result<heif_item_id> HeifFile::add_infe_uri(const char* item_uri_type, const uint8_t* data, size_t size)
+{
+  Result<heif_item_id> result;
+
+  // create an infe box describing what kind of data we are storing (this also creates a new ID)
+
+  auto infe_box = add_new_infe_box("uri ");
+  infe_box->set_hidden_item(true);
+  infe_box->set_item_uri_type(item_uri_type);
+
+  heif_item_id metadata_id = infe_box->get_item_ID();
+  result.value = metadata_id;
+
+  set_item_data(infe_box, data, size, heif_metadata_compression_off);
+
+  return result;
+}
+
+
+Error HeifFile::set_item_data(const std::shared_ptr<Box_infe>& item, const uint8_t* data, size_t size, heif_metadata_compression compression)
+{
+  // --- metadata compression
+
+  if (compression == heif_metadata_compression_auto) {
+    compression = heif_metadata_compression_off; // currently, we don't use header compression by default
+  }
+
+  // only set metadata compression for MIME type data which has 'content_encoding' field
+  if (compression != heif_metadata_compression_off &&
+      item->get_item_type() != "mime") {
+    // TODO: error, compression not supported
+  }
+
+
+  std::vector<uint8_t> data_array;
+  if (compression == heif_metadata_compression_zlib) {
+#if HAVE_ZLIB
+    data_array = compress_zlib((const uint8_t*) data, size);
+    item->set_content_encoding("compress_zlib");
+#else
+    return Error(heif_error_Unsupported_feature,
+                 heif_suberror_Unsupported_header_compression_method);
+#endif
+  }
+  else if (compression == heif_metadata_compression_deflate) {
+#if HAVE_ZLIB
+    data_array = compress_deflate((const uint8_t*) data, size);
+    item->set_content_encoding("compress_zlib");
+#else
+    return Error(heif_error_Unsupported_feature,
+                 heif_suberror_Unsupported_header_compression_method);
+#endif
+  }
+  // TODO: brotli
+  else {
+    // uncompressed data, plain copy
+
+    data_array.resize(size);
+    memcpy(data_array.data(), data, size);
+  }
+
+  // copy the data into the file, store the pointer to it in an iloc box entry
+
+  append_iloc_data(item->get_item_ID(), data_array);
+
+  return Error::Ok;
+}
+
+
+Error HeifFile::set_precompressed_item_data(const std::shared_ptr<Box_infe>& item, const uint8_t* data, size_t size, std::string content_encoding)
+{
+  // only set metadata compression for MIME type data which has 'content_encoding' field
+  if (!content_encoding.empty() &&
+      item->get_item_type() != "mime") {
+    // TODO: error, compression not supported
+  }
+
+
+  std::vector<uint8_t> data_array;
+  data_array.resize(size);
+  memcpy(data_array.data(), data, size);
+
+  item->set_content_encoding(content_encoding);
+
+  // copy the data into the file, store the pointer to it in an iloc box entry
+
+  append_iloc_data(item->get_item_ID(), data_array);
+
+  return Error::Ok;
+}
+
+
 void HeifFile::append_iloc_data(heif_item_id id, const std::vector<uint8_t>& nal_packets, uint8_t construction_method)
 {
   m_iloc_box->append_data(id, nal_packets, construction_method);
@@ -1216,7 +1797,7 @@ void HeifFile::set_auxC_property(heif_item_id id, const std::string& type)
   auto auxC = std::make_shared<Box_auxC>();
   auxC->set_aux_type(type);
 
-  int index = m_ipco_box->append_child_box(auxC);
+  int index = m_ipco_box->find_or_append_child_box(auxC);
 
   m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{true, uint16_t(index + 1)});
 }
@@ -1226,7 +1807,7 @@ void HeifFile::set_color_profile(heif_item_id id, const std::shared_ptr<const co
   auto colr = std::make_shared<Box_colr>();
   colr->set_color_profile(profile);
 
-  int index = m_ipco_box->append_child_box(colr);
+  int index = m_ipco_box->find_or_append_child_box(colr);
   m_ipma_box->add_property_for_item_ID(id, Box_ipma::PropertyAssociation{false, uint16_t(index + 1)});
 }
 
