@@ -40,6 +40,7 @@ struct encoder_struct_opj
 {
   int quality = 70;
   heif_chroma chroma = heif_chroma_undefined;
+  opj_cparameters_t parameters;
 
   // --- output
 
@@ -124,6 +125,7 @@ struct heif_error opj_new_encoder(void** encoder_out)
   *encoder_out = encoder;
 
   opj_set_default_parameters(encoder);
+  opj_set_default_encoder_parameters(&(encoder->parameters));
 
   return heif_error_ok;
 }
@@ -156,13 +158,17 @@ struct heif_error opj_get_parameter_quality(void* encoder_raw, int* quality)
   return heif_error_ok;
 }
 
-struct heif_error opj_set_parameter_lossless(void* encoder, int lossless)
+struct heif_error opj_set_parameter_lossless(void* encoder_raw, int lossless)
 {
+  auto* encoder = (struct encoder_struct_opj*) encoder_raw;
+  encoder->parameters.irreversible = lossless ? 0 : 1;
   return heif_error_ok;
 }
 
-struct heif_error opj_get_parameter_lossless(void* encoder, int* lossless)
+struct heif_error opj_get_parameter_lossless(void* encoder_raw, int* lossless)
 {
+  auto* encoder = (struct encoder_struct_opj*) encoder_raw;
+  *lossless = (encoder->parameters.irreversible == 0);
   return heif_error_ok;
 }
 
@@ -248,22 +254,25 @@ struct heif_error opj_get_parameter_string(void* encoder_raw, const char* name,
 {
   struct encoder_struct_opj* encoder = (struct encoder_struct_opj*) encoder_raw;
 
-  switch (encoder->chroma) {
-    case heif_chroma_420:
-      save_strcpy(value, value_size, "420");
-      break;
-    case heif_chroma_422:
-      save_strcpy(value, value_size, "422");
-      break;
-    case heif_chroma_444:
-      save_strcpy(value, value_size, "444");
-      break;
-    case heif_chroma_undefined:
-      save_strcpy(value, value_size, "undefined");
-      break;
-    default:
-      assert(false);
-      return heif_error_invalid_parameter_value;
+  if (strcmp(name, kParam_chroma) == 0) {
+    switch (encoder->chroma) {
+      case heif_chroma_420:
+        save_strcpy(value, value_size, "420");
+        break;
+      case heif_chroma_422:
+        save_strcpy(value, value_size, "422");
+        break;
+      case heif_chroma_444:
+        save_strcpy(value, value_size, "444");
+        break;
+      case heif_chroma_undefined:
+        save_strcpy(value, value_size, "undefined");
+        break;
+      default:
+        assert(false);
+        return heif_error_invalid_parameter_value;
+      }
+      return heif_error_ok;
   }
 
   return heif_error_unsupported_parameter;
@@ -358,12 +367,10 @@ static heif_error generate_codestream(opj_image_t* image, struct encoder_struct_
 {
   heif_error error;
   OPJ_BOOL success;
-  opj_cparameters_t parameters;
-  opj_set_default_encoder_parameters(&parameters);
 
-  parameters.cp_disto_alloc = 1;
-  parameters.tcp_numlayers = 1;
-  parameters.tcp_rates[0] = (float)(1 + (100 - encoder->quality)/2);
+  encoder->parameters.cp_disto_alloc = 1;
+  encoder->parameters.tcp_numlayers = 1;
+  encoder->parameters.tcp_rates[0] = (float)(1 + (100 - encoder->quality)/2);
 
 #if 0
   //Insert a human readable comment into the codestream
@@ -382,8 +389,9 @@ static heif_error generate_codestream(opj_image_t* image, struct encoder_struct_
   //OPJ_CODEC_JP2 - Generate the entire jp2 file (which contains a codestream)
   OPJ_CODEC_FORMAT codec_format = OPJ_CODEC_J2K;
   opj_codec_t* codec = opj_create_compress(codec_format);
-  success = opj_setup_encoder(codec, &parameters, image);
+  success = opj_setup_encoder(codec, &(encoder->parameters), image);
   if (!success) {
+    opj_destroy_codec(codec);
     error = {heif_error_Encoding_error, heif_suberror_Unspecified, "Failed to setup OpenJPEG encoder"};
     return error;
   }
@@ -393,6 +401,7 @@ static heif_error generate_codestream(opj_image_t* image, struct encoder_struct_
   size_t bufferSize = 64 * 1024; // 64k
   opj_stream_t* stream = opj_stream_create(bufferSize, false /* read only mode */);
   if (stream == NULL) {
+    opj_destroy_codec(codec);
     error = {heif_error_Encoding_error, heif_suberror_Unspecified, "Failed to create opj_stream_t"};
     return error;
   }
@@ -411,22 +420,31 @@ static heif_error generate_codestream(opj_image_t* image, struct encoder_struct_
 
   success = opj_start_compress(codec, image, stream);
   if (!success) {
+    opj_stream_destroy(stream);
+    opj_destroy_codec(codec);
     error = {heif_error_Encoding_error, heif_suberror_Unspecified, "Failed opj_start_compress()"};
     return error;
   }
 
   success = opj_encode(codec, stream);
   if (!success) {
+    opj_stream_destroy(stream);
+    opj_destroy_codec(codec);
     error = {heif_error_Encoding_error, heif_suberror_Unspecified, "Failed opj_encode()"};
     return error;
   }
 
   success = opj_end_compress(codec, stream);
   if (!success) {
+    opj_stream_destroy(stream);
+    opj_destroy_codec(codec);
     error = {heif_error_Encoding_error, heif_suberror_Unspecified, "Failed opj_end_compress()"};
     return error;
   }
 
+  opj_stream_destroy(stream);
+  opj_destroy_codec(codec);
+
   return heif_error_ok;
 }
 
@@ -527,6 +545,8 @@ struct heif_error opj_encode_image(void* encoder_raw, const struct heif_image* i
   //Encodes the image into a 'codestream' which is stored in the 'encoder' variable
   err = generate_codestream(opj_image, encoder);
 
+  opj_image_destroy(opj_image);
+
   return err;
 }
 
